import os
import re
import json
import abc
import warnings
from typing import MutableMapping, List, Union
from functools import reduce
from enum import Enum

import pandas as pd
import numpy as np
from scipy import sparse
import loompy as lp

from loomxpy import __DEBUG__
from loomxpy._specifications import (
    ProjectionMethod,
    LoomXMetadataEmbedding,
    LoomXMetadataClustering,
    LoomXMetadataCluster,
    LoomXMetadataClusterMarkerMetric,
)
from loomxpy._s7 import S7
from loomxpy._errors import BadDTypeException
from loomxpy._hooks import WithInitHook
from loomxpy._matrix import DataMatrix
from loomxpy.utils import df_to_named_matrix, compress_encode


def custom_formatwarning(msg, *args, **kwargs):
    # ignore everything except the message
    return str(msg) + "\n"


warnings.formatwarning = custom_formatwarning


##########################################
# MODES                                  #
##########################################


class ModeType(Enum):
    NONE = "_"
    RNA = "rna"


class Mode(S7):
    def __init__(self, mode_type: ModeType, data_matrix: DataMatrix):
        """
        constructor for Mode
        """
        self._mode_type = mode_type
        # Data Matrix
        self._data_matrix = data_matrix
        # Global
        self._global_attrs = GlobalAttributes(mode=self)
        # Features
        self._feature_attrs = FeatureAttributes(mode=self)
        self._fa_annotations = FeatureAnnotationAttributes(mode=self)
        self._fa_metrics = FeatureMetricAttributes(mode=self)
        # Observations
        self._observation_attrs = ObservationAttributes(mode=self)
        self._oa_annotations = ObservationAnnotationAttributes(mode=self)
        self._oa_metrics = ObservationMetricAttributes(mode=self)
        self._oa_embeddings = ObservationEmbeddingAttributes(mode=self)
        self._oa_clusterings = ObservationClusteringAttributes(mode=self)

    @property
    def X(self):
        return self._data_matrix

    @property
    def g(self):
        return self._global_attrs

    @property
    def f(self):
        return self._feature_attrs

    @property
    def o(self):
        return self._observation_attrs

    def export(
        self,
        filename: str,
        output_format: str,
        title: str = None,
        genome: str = None,
        compress_metadata: bool = False,
        cluster_marker_metrics: List[LoomXMetadataClusterMarkerMetric] = [
            {
                "accessor": "avg_logFC",
                "name": "Avg. logFC",
                "description": f"Average log fold change from Wilcoxon test",
                "threshold": 0,
                "threshold_method": "lte_or_gte",  # lte, lt, gte, gt, lte_or_gte, lte_and_gte
            },
            {
                "accessor": "pval",
                "name": "Adjusted P-Value",
                "description": f"Adjusted P-Value from Wilcoxon test",
                "threshold": 0.05,
                "threshold_method": "lte",  # lte, lt, gte, gt, lte_or_gte, lte_and_gte
            },
        ],
    ):
        """
        Export this LoomX object to Loom file

        Parameters
        ---------
        cluster_marker_metrics: dict, optional
            List of dict (ClusterMarkerMetric) containing metadata of each metric available for the cluster markers.
            Expects each metric to be of type float.
        Return
        ------
        None
        """
        if output_format == "scope_v1":
            #
            _feature_names = self._data_matrix._feature_names
            # Init
            _row_attrs: MutableMapping = {}
            _col_attrs: MutableMapping = {}
            _global_attrs: MutableMapping = {
                "title": os.path.splitext(os.path.basename(filename))[0]
                if title is None
                else title,
                "MetaData": {
                    "annotations": [],
                    "metrics": [],
                    "embeddings": [],
                    "clusterings": [],
                },
                "Genome": genome,
            }
            # Add row attributes (in Loom specifications)
            for _attr_key, _attr in self._feature_attrs:
                _row_attrs[_attr_key] = _attr.values

            # Add columns attributes (in Loom specifications)
            _default_embedding = None
            _embeddings_X = pd.DataFrame(index=self._data_matrix._observation_names)
            _embeddings_Y = pd.DataFrame(index=self._data_matrix._observation_names)
            _clusterings = pd.DataFrame(index=self._data_matrix._observation_names)

            for _attr_key, _attr in self._observation_attrs:
                if _attr.attr_type.value == AttributeType.ANNOTATION.value:
                    # Categorical not valid, ndarray is required
                    _col_attrs[_attr_key] = np.asarray(_attr.values)
                    _global_attrs["MetaData"]["annotations"].append(
                        {
                            "name": _attr_key,
                            "values": list(
                                map(
                                    lambda x: x.item()
                                    if type(x).__module__ == "numpy"
                                    else x,
                                    sorted(
                                        np.unique(_attr.values),
                                        reverse=False,
                                    ),
                                )
                            ),
                        }
                    )
                if _attr.attr_type.value == AttributeType.METRIC.value:
                    _col_attrs[_attr_key] = np.asarray(_attr.values)
                    _global_attrs["MetaData"]["metrics"].append({"name": _attr_key})

                if _attr.attr_type.value == AttributeType.EMBEDDING.value:
                    _attr: EmbeddingAttribute
                    _data = _attr.data.iloc[:, 0:2]
                    _data.columns = ["_X", "_Y"]

                    # Number of embeddings (don't count the default embedding since this will be use to determine the id of the embedding)
                    _num_embeddings = len(
                        list(
                            filter(
                                lambda x: int(x["id"]) != -1,
                                _global_attrs["MetaData"]["embeddings"],
                            )
                        )
                    )
                    _embedding_id = (
                        _attr.id
                        if _attr.id is not None
                        else (
                            -1
                            if _attr._default
                            else 0
                            if _num_embeddings == 0
                            else _num_embeddings + 1
                        )
                    )
                    _embeddings_X = pd.merge(
                        _embeddings_X,
                        _data["_X"]
                        .to_frame()
                        .rename(columns={"_X": str(_embedding_id)})
                        .astype("float32"),
                        left_index=True,
                        right_index=True,
                    )
                    _embeddings_Y = pd.merge(
                        _embeddings_Y,
                        _data["_Y"]
                        .to_frame()
                        .rename(columns={"_Y": str(_embedding_id)})
                        .astype("float32"),
                        left_index=True,
                        right_index=True,
                    )
                    _global_attrs["MetaData"]["embeddings"].append(
                        {
                            "id": str(
                                _embedding_id
                            ),  # TODO: type not consistent with clusterings
                            "name": _attr.name,
                        }
                    )

                    if _attr.default:
                        _default_embedding = _data

                if _attr.attr_type.value == AttributeType.CLUSTERING.value:
                    _attr: ClusteringAttribute
                    if _attr.name is None:
                        raise Exception(
                            f"The clustering with key '{_attr.key}' does not have a name. This is required when exporting to SCope."
                        )

                    # Clustering
                    _col_name = (
                        _attr.data.columns[0]
                        if isinstance(_attr, pd.DataFrame)
                        else _attr.name
                    )
                    _num_clusterings = len(_global_attrs["MetaData"]["clusterings"])
                    _clustering_id = (
                        0 if _num_clusterings == 0 else _num_clusterings + 1
                    )
                    _clustering_data = (
                        _attr.data.rename(columns={_col_name: str(_clustering_id)})
                        if isinstance(_attr.data, pd.DataFrame)  # pd.DataFrame
                        else _attr.data.rename(str(_clustering_id))  # pd.Series
                    )
                    _clusterings = pd.merge(
                        _clusterings,
                        _clustering_data,
                        left_index=True,
                        right_index=True,
                    )
                    _clustering_md = LoomXMetadataClustering.from_dict(
                        {
                            "id": _clustering_id,
                            **_attr.metadata.to_dict(),
                        }
                    ).to_dict()

                    # Markers
                    # Dictionary of DataFrame (value) containing the the values of the different metric (key) across features (rows) and for each cluster (columns)
                    _cluster_markers_dict = {}

                    if cluster_marker_metrics:

                        has_cluster_markers = [
                            cluster.markers is not None
                            for cluster in _attr._metadata.clusters
                        ]
                        if not all(has_cluster_markers):
                            continue

                        # Init DataFrame mask of genes representing markers
                        cluster_markers = pd.DataFrame(
                            index=_feature_names,
                            columns=[str(x.id) for x in _attr._metadata.clusters],
                        ).fillna(0, inplace=False)

                        # Init DataFrame containing metric valuess
                        _cluster_marker_metric: LoomXMetadataClusterMarkerMetric
                        for _cluster_marker_metric in cluster_marker_metrics:
                            _cluster_markers_dict[
                                _cluster_marker_metric["accessor"]
                            ] = pd.DataFrame(
                                index=_feature_names,
                                columns=[str(x.id) for x in _attr._metadata.clusters],
                            ).fillna(
                                0, inplace=False
                            )

                        _cluster: LoomXMetadataCluster
                        for _cluster in _attr._metadata.clusters:

                            _features_df = pd.Series(
                                _cluster.markers.index.values,
                                index=_cluster.markers.index.values,
                            )

                            # Dictionary of Series (value) containing the values of the different metric (key) for the current cluster
                            _cluster_marker_metric_values_dict = {}
                            # Dictionary of Series (value) containing a boolean mask of the features that pass the filter criteria for the different metrics (key)
                            _cluster_marker_metric_masks_dict = {}

                            _cluster_marker_metric: LoomXMetadataClusterMarkerMetric
                            for _cluster_marker_metric in cluster_marker_metrics:

                                # Check if metric exists in markers table
                                if (
                                    _cluster_marker_metric["accessor"]
                                    not in _cluster.markers.columns
                                ):
                                    raise Exception(
                                        f"The cluster_marker_metrics argument was not properly defined. Missing {_cluster_marker_metric['accessor']} metric in the markers table. Available columns in markers table are f{''.join(_cluster.markers.columns)}."
                                    )

                                cluster_marker_metric_values = pd.Series(
                                    _cluster.markers[
                                        _cluster_marker_metric["accessor"]
                                    ].values,
                                    index=_cluster.markers.index.values,
                                ).astype(float)

                                if pd.isnull(cluster_marker_metric_values).any():
                                    raise Exception(
                                        f"NaN detected in {_cluster_marker_metric['accessor']} metric column of the markers table"
                                    )

                                if _cluster_marker_metric["threshold_method"] == "lte":
                                    feature_mask = (
                                        cluster_marker_metric_values
                                        < _cluster_marker_metric["threshold"]
                                    )
                                elif _cluster_marker_metric["threshold_method"] == "lt":
                                    feature_mask = (
                                        cluster_marker_metric_values
                                        <= _cluster_marker_metric["threshold"]
                                    )
                                elif (
                                    _cluster_marker_metric["threshold_method"]
                                    == "lte_or_gte"
                                ):
                                    feature_mask = np.logical_and(
                                        np.logical_or(
                                            cluster_marker_metric_values
                                            >= _cluster_marker_metric["threshold"],
                                            cluster_marker_metric_values
                                            <= -_cluster_marker_metric["threshold"],
                                        ),
                                        np.isfinite(cluster_marker_metric_values),
                                    )
                                else:
                                    raise Exception(
                                        "The given threshold method is not implemented"
                                    )
                                _cluster_marker_metric_masks_dict[
                                    _cluster_marker_metric["accessor"]
                                ] = feature_mask
                                _cluster_marker_metric_values_dict[
                                    _cluster_marker_metric["accessor"]
                                ] = cluster_marker_metric_values

                            # Create a new cluster marker mask based on all feature mask generated using each metric
                            cluster_marker_metrics_mask = np.logical_or.reduce(
                                [
                                    v
                                    for _, v in _cluster_marker_metric_masks_dict.items()
                                ]
                            )

                            marker_names = _features_df[cluster_marker_metrics_mask]

                            # Get a marker mask along all features in the matrix
                            marker_genes_along_data_mask = np.in1d(
                                _feature_names, marker_names
                            )
                            marker_genes_along_data = cluster_markers.index[
                                marker_genes_along_data_mask
                            ]

                            # Populate the marker mask
                            markers_df = pd.DataFrame(
                                1, index=marker_names, columns=["is_marker"]
                            )
                            cluster_markers.loc[
                                marker_genes_along_data_mask, str(_cluster.id)
                            ] = markers_df["is_marker"][marker_genes_along_data]

                            if pd.isnull(cluster_markers[str(_cluster.id)]).any():
                                raise Exception(
                                    f"NaN detected in markers DataFrame of cluster {_cluster.id}."
                                )

                            # Populate the marker nmetrics
                            for _cluster_marker_metric in _cluster_markers_dict.keys():
                                _metric_df = pd.DataFrame(
                                    _cluster_marker_metric_values_dict[
                                        _cluster_marker_metric
                                    ][cluster_marker_metrics_mask],
                                    index=marker_names,
                                    columns=[_cluster_marker_metric],
                                )
                                _cluster_markers_dict[_cluster_marker_metric].loc[
                                    marker_genes_along_data_mask, str(_cluster.id)
                                ] = _metric_df[_cluster_marker_metric][
                                    marker_genes_along_data
                                ]

                                if pd.isnull(
                                    _cluster_markers_dict[_cluster_marker_metric][
                                        str(_cluster.id)
                                    ]
                                ).any():
                                    raise Exception(
                                        f"NaN detected in markers metric {_cluster_marker_metric['accessor']} DataFrame of cluster {_cluster.id}."
                                    )

                        # Add the required global metadata for markes to be visualized in SCope
                        # Encapsule with mixin to avoid properties not required by gRPC

                        _clustering_md["clusterMarkerMetrics"] = [
                            LoomXMetadataClusterMarkerMetric.from_dict(cmm).to_dict()
                            for cmm in cluster_marker_metrics
                        ]

                    _global_attrs["MetaData"]["clusterings"].append(_clustering_md)

                    # Convert all markers related data to Loom compliant format
                    _row_attrs[
                        f"ClusterMarkers_{str(_clustering_id)}"
                    ] = df_to_named_matrix(cluster_markers)

                    for _cluster_marker_metric in _cluster_markers_dict.keys():
                        _row_attrs[
                            f"ClusterMarkers_{str(_clustering_id)}_{_cluster_marker_metric}"
                        ] = df_to_named_matrix(
                            _cluster_markers_dict[_cluster_marker_metric].astype(float)
                        )

            _row_attrs["Gene"] = np.asarray(_feature_names)
            _col_attrs["CellID"] = np.asarray(self._data_matrix._observation_names)

            # If no default embedding, use the first embedding as default
            if _default_embedding is None:
                _col_attrs["Embedding"] = df_to_named_matrix(
                    df=pd.DataFrame(
                        {
                            "_X": _embeddings_X["0"].values,
                            "_Y": _embeddings_Y["0"].values,
                        }
                    )
                )
                _embeddings_X.insert(
                    loc=0, column="-1", value=_embeddings_X["0"].values
                )
                _embeddings_Y.insert(
                    loc=0, column="-1", value=_embeddings_Y["0"].values
                )
                _md_first_embedding = list(
                    filter(
                        lambda x: x["id"] == "0",
                        _global_attrs["MetaData"]["embeddings"],
                    )
                )[0]
                _global_attrs["MetaData"]["embeddings"] = [
                    {"id": "-1", "name": _md_first_embedding["name"]}
                ] + list(
                    filter(
                        lambda x: x["id"] != "0",
                        _global_attrs["MetaData"]["embeddings"],
                    )
                )
            else:
                _col_attrs["Embedding"] = df_to_named_matrix(df=_default_embedding)

            _col_attrs["Embeddings_X"] = df_to_named_matrix(df=_embeddings_X)
            _col_attrs["Embeddings_Y"] = df_to_named_matrix(df=_embeddings_Y)
            _col_attrs["Clusterings"] = df_to_named_matrix(
                df=_clusterings.astype(np.int16)
            )

            _global_attrs["MetaData"] = json.dumps(_global_attrs["MetaData"])

            if compress_metadata:
                _global_attrs["MetaData"] = compress_encode(
                    value=_global_attrs["MetaData"]
                )

            for _ga_key, _ga_value in self._global_attrs:
                _global_attrs[_ga_key] = _ga_value

            lp.create(
                filename=filename,
                layers=self._data_matrix._data_matrix.transpose(),
                row_attrs=_row_attrs,
                col_attrs=_col_attrs,
                file_attrs=_global_attrs,
            )
            print("INFO: LoomX successfully exported to SCope-compatible loom file.")

        else:
            raise Exception(
                f"Cannot export LoomX to the given output format '{output_format}'. Invalid output format"
            )


class Modes(MutableMapping[str, Mode], metaclass=WithInitHook):
    def __init__(self):
        """"""
        self._keys: List[str] = []
        self._mode_types = [item.value for item in ModeType]
        # Implemented modes (used here mainly for typing purposes)
        self.rna: Mode = None

    def __setattr__(self, name, value):
        if not hasattr(self, "_initialized"):
            if __DEBUG__:
                print(f"DEBUG: constructor call: set attr with name {name}")
            super().__setattr__(name, value)
        else:
            self.__setitem__(name=name, value=value)

    def __delattr__(self, name: str):
        self._keys.remove(name)
        super().__delattr__(name)

    def __iter__(self):
        """"""
        raise NotImplementedError

    def __len__(self):
        """"""
        raise NotImplementedError

    def __delitem__(self, name) -> None:
        """"""
        self.__delattr__(name)

    def __getitem__(self, name) -> Mode:
        """"""
        return getattr(self, name)

    def _validate_key(self, key):
        _key = None
        # FIXME: Fix for editable mode (get called with name=__class__)
        if key.startswith("__"):
            return

        if key.startswith("_"):
            raise Exception(
                f"Cannot add Mode with key {key}. Not a valid key. Expects key not to start with an underscore ('_')."
            )

        if not isinstance(key, ModeType) and not str:
            raise Exception("Not a valid ModeType.")

        if isinstance(key, ModeType):
            if key in self._mode_types:
                _key = key
            else:
                raise Exception("Not a valid ModeType.")

        if isinstance(key, str):
            if key in self._mode_types:
                _key = key
            else:
                raise Exception(
                    f"Not a valid ModeType: {key}. Choose one ModeType from {', '.join(self._mode_types)}"
                )
        return _key

    @staticmethod
    def _validate_value(value):
        if not type(value) in [tuple, np.matrix, pd.DataFrame]:
            raise Exception(
                f"""
Got {type(value)} but expecting either:
- Tuple: (<SciPy CSR matrix>, <feature names>, <observation names>) or
- Tuple: (<NumPy 2D matrix>, <feature names>, <observation names>) or
- pandas.DataFrame
                """
            )

        if isinstance(value, sparse.csr_matrix):
            raise Exception(
                "If your data matrix is a SciPy CSR matrix, use tuple: (<SciPy CSR matrix>, <feature names>, <observation names>)."
            )

        if isinstance(value, sparse.csc_matrix):
            raise Exception(
                "If your data matrix is a SciPy CSC matrix, use tuple: (<SciPy CSC matrix>, <feature names>, <observation names>)."
            )

        if isinstance(value, np.matrix):
            raise Exception(
                "If your data matrix is a NumPy 2D matrix, use tuple: (<NumPy 2D matrix>, <feature names>, <observation names>)."
            )

        if isinstance(value, tuple):
            if len(value) != 3:
                raise Exception(
                    "If your data matrix is a NumPy 2D matrix or SciPy CSR matrix, use tuple: (<NumPy 2D matrix | SciPy CSR matrix>, <feature names>, <observation names>)."
                )

    def __setitem__(self, name, value) -> None:
        """"""
        if __DEBUG__:
            print(f"DEBUG: instance call: set attr with name {name}")

        # FIXME: Fix for editable mode (get called with name=__class__)
        if name.startswith("__"):
            return
        print(f"INFO: Adding new {name} mode")
        _key = self._validate_key(key=name)
        Modes._validate_value(value=value)

        _mode = None
        _data_matrix = None

        if isinstance(value, tuple):
            _matrix, _feature_names, _observation_names = value
            _data_matrix = DataMatrix(
                data_matrix=_matrix,
                feature_names=_feature_names,
                observation_names=_observation_names,
            )

        if isinstance(value, pd.DataFrame):
            _data_matrix = DataMatrix(
                data_matrix=value.values,
                feature_names=value.columns,
                observation_names=value.index,
            )

        if _data_matrix is None:
            raise Exception("Invalid type of the given data natrix.")

        _mode = Mode(mode_type=ModeType(_key), data_matrix=_data_matrix)
        if _key not in self._keys:
            self._keys.append(_key)
        super().__setattr__(_key, _mode)

    def __repr__(self) -> str:
        _mode_keys = f"{', '.join(self._keys)}" if len(self._keys) > 0 else "none"
        return f"Modalities: {_mode_keys}"


##########################################
# ATTRIBUTES                             #
##########################################


class GlobalAttributes(MutableMapping[str, str], metaclass=WithInitHook):
    def __init__(self, mode: Mode):
        """"""
        self._keys: List[str] = []
        # Implemented modes (used here mainly for typing purposes)
        self.rna: Mode = mode

    def __setattr__(self, name, value):
        if not hasattr(self, "_initialized"):
            if __DEBUG__:
                print(f"DEBUG: constructor call: set attr with name {name}")
            super().__setattr__(name, value)
        else:
            self.__setitem__(name=name, value=value)

    def __delattr__(self, name: str):
        self._keys.remove(name)
        super().__delattr__(name)

    def __iter__(self):
        """"""
        return iter(AttributesIterator(self))

    def __len__(self):
        """"""
        return len(self._keys)

    def __delitem__(self, name: str) -> None:
        """"""
        self.__delattr__(name)

    def __getitem__(self, name: str) -> Mode:
        """"""
        return getattr(self, name)

    def __setitem__(self, name: str, value: str) -> None:
        """"""
        # FIXME: Fix for editable mode (get called with name=__class__)
        if name.startswith("__"):
            return

        if not isinstance(name, str):
            raise Exception("Not a valid key for GlobalAttribute.")

        if not isinstance(value, str):
            raise Exception("Not a valid value for GlobalAttribute.")

        self._add_key(key=name)
        super().__setattr__(name, value)

    def get_attribute(self, key: str):
        """"""
        return super().__getattribute__(key)

    def __repr__(self) -> str:
        _keys = f"{', '.join(self._keys)}" if len(self._keys) > 0 else "none"
        return f"Global attributes: {_keys}"

    def _add_key(self, key: str):
        if key not in self._keys:
            self._keys.append(key)


class GlobalAttributesIterator:

    """Class to implement an iterator of Attributes """

    def __init__(self, attrs: GlobalAttributes):
        self._attrs = attrs

    def __iter__(self):
        self._n = 0
        return self

    def __next__(self):
        if self._n < len(self._attrs._keys):
            current_key = self._attrs._keys[self._n]
            self._n += 1
            return current_key, self._attrs.get_attribute(current_key)
        else:
            raise StopIteration


class Axis(Enum):
    OBSERVATIONS = 0
    FEATURES = 1


class AttributeType(Enum):
    ANNOTATION = 0
    METRIC = 1
    EMBEDDING = 2
    CLUSTERING = 3


class Attribute:
    def __init__(
        self,
        key: str,
        mode_type: ModeType,
        attr_type: AttributeType,
        axis: Axis,
        data,
        name: str = None,
        description: str = None,
    ):
        """"""
        self._key = key
        self._mode_type = mode_type
        self._attr_type = attr_type
        self._axis = axis
        self._data = data
        self._name = name
        self._description = description

    @property
    def key(self):
        return self._key

    @property
    def mode_type(self):
        return self._mode_type

    @property
    def attr_type(self):
        return self._attr_type

    @property
    def axis(self):
        return self._axis

    @property
    def data(self):
        return self._data

    @property
    def values(self):
        if isinstance(self._data, pd.DataFrame):
            _col_name = self._data.columns[0]
            return self._data[_col_name].values
        if isinstance(self._data, pd.Series):
            return self._data
        raise Exception(f"Cannot get values from Attribute with key {self._key}")

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def description(self):
        return self._description

    def __repr__(self):
        """"""
        return f"""
key: {self._key}
mode: {self._mode_type}
type: {self._attr_type}
name: {self._name}
description: {self._description}
        """


class Attributes(MutableMapping[str, Attribute], metaclass=WithInitHook):
    def __init__(self, mode: Mode, axis: Axis, is_proxy: bool = False, **kwargs):
        """"""
        self._keys: List[str] = []
        self._mode = mode
        self._mode_type = mode._mode_type if mode is not None else ModeType.NONE
        self._axis = axis
        self._is_proxy = is_proxy
        self._attr_type = kwargs["attr_type"] if "attr_type" in kwargs else False
        self._is_multi = True if "is_multi" in kwargs and kwargs["is_multi"] else False

    def __getattribute__(self, key):
        """"""
        if not super().__getattribute__("_is_proxy"):
            if key in super().__getattribute__("_keys"):
                return super().__getattribute__(key).data
            return super().__getattribute__(key)
        else:
            """
            This is a proxy. Override __getattribute__ of Attributes class
            """
            if key in super().__getattribute__("_keys"):
                if super().__getattribute__("_axis") == Axis.FEATURES:
                    return super().__getattribute__("_mode")._feature_attrs[key]
                elif super().__getattribute__("_axis") == Axis.OBSERVATIONS:
                    return super().__getattribute__("_mode")._observation_attrs[key]
                else:
                    raise Exception("Invalid axis.")
            return super().__getattribute__(key)

    def __setattr__(self, name, value):
        if not hasattr(self, "_initialized"):
            if __DEBUG__:
                print(f"DEBUG: constructor call: set attr with name {name}")
            super().__setattr__(name, value)
        else:
            self.__setitem__(name=name, value=value)

    def __delattr__(self, name):
        self._keys.remove(name)
        super().__delattr__(name)

    def __delitem__(self, key):
        """"""
        self.__delattr__(key)

    def __getitem__(self, key):
        """"""
        return super().__getattribute__(key).data

    def __iter__(self):
        """"""
        return iter(AttributesIterator(self))

    def __len__(self):
        """"""
        return len(self._keys)

    def _add_key(self, key: str):
        if key not in self._keys:
            self._keys.append(key)

    def _add_item(self, key: str, value: Attribute) -> None:
        self._add_key(key=key)
        super().__setattr__(key, value)

    def _add_item_by_value(self, value: Attribute):
        self._add_key(key=value.key)
        super().__setattr__(value.key, value)

    def get_attribute(self, key):
        """"""
        return super().__getattribute__(key)

    @abc.abstractclassmethod
    def __setitem__(self, name, value):
        """"""
        raise NotImplementedError

    def _validate_key(self, key):
        # FIXME: Fix for editable mode (get called with name=__class__)
        if key.startswith("__"):
            return

        if key.startswith("_"):
            raise Exception(
                f"Cannot add attribute with key {key}. Not a valid key. Expects key not to start with an underscore ('_')."
            )
        if not isinstance(key, str):
            raise Exception(
                f"Cannot add attribute with key of type ({type(key).__name__}) to {type(self).__name__}. Not a valid key. Expects key of type str."
            )
        # Print a warning in key contains characters not allowed. If any present, this will prevent the user to use dot notation. Brackets access will work.
        pattern = "^[a-zA-Z0-9_]+$"
        if not re.match(pattern, key):
            warnings.warn(
                f"The key '{key}' won't be accessible using the dot notation (containing special characters other than '_')",
            )

    def _validate_value(self, value):
        if not isinstance(value, pd.DataFrame) and not isinstance(value, pd.Series):
            raise Exception(
                f"Cannot add attribute of type {type(value).__name__} to {type(self).__name__}. Expects a pandas.DataFrame or a pandas.Series."
            )

        if (
            isinstance(value, pd.DataFrame)
            and not self._is_multi
            and value.shape[1] > 1
        ):
            raise Exception(
                f"Cannot add attribute of shape {value.shape[1]}. Currently, allows only {type(value).__name__} with maximally 1 feature (i.e.: column)."
            )


class AttributesIterator:

    """Class to implement an iterator of Attributes """

    def __init__(self, attrs: Attributes):
        self._attrs = attrs

    def __iter__(self):
        self._n = 0
        return self

    def __next__(self):
        if self._n < len(self._attrs._keys):
            current_key = self._attrs._keys[self._n]
            self._n += 1
            return current_key, self._attrs.get_attribute(current_key)
        else:
            raise StopIteration


##########################################
# ATTRIBUTE TYPES                        #
##########################################


class AnnotationAttributes(Attributes):
    def __init__(self, mode: Mode, axis: Axis, is_proxy: bool = False, **kwargs):
        """"""
        super().__init__(
            mode=mode,
            axis=axis,
            is_proxy=is_proxy,
            attr_type=AttributeType.ANNOTATION,
            **kwargs,
        )

    def _validate_value(self, value: Union[pd.DataFrame, pd.Series], **kwargs):
        if __DEBUG__:
            print(f"DEBUG: _validate_value ({type(self).__name__})")
        super()._validate_value(value=value)
        _force_conversion_to_categorical = (
            kwargs["force_conversion_to_categorical"]
            if "force_conversion_to_categorical" in kwargs
            else False
        )
        # Do some checks and processing for attribute of type ANNOTATION
        if not _force_conversion_to_categorical:
            if (
                isinstance(value, pd.DataFrame)
                and not all(value.apply(pd.api.types.is_categorical_dtype))
                and not all(value.apply(pd.api.types.is_bool_dtype))
            ) or (
                isinstance(value, pd.Series)
                and not pd.api.types.is_categorical_dtype(arr_or_dtype=value)
                and not pd.api.types.is_bool_dtype(arr_or_dtype=value)
            ):
                _dtype = (
                    value.infer_objects().dtypes[0]
                    if isinstance(value, pd.DataFrame)
                    else value.infer_objects().dtype
                )
                raise BadDTypeException(
                    f"Expects value to be categorical or bool but its dtype is {_dtype}. You can force the conversion to categorical by using <loomx-instance>.modes.<mode>.annotations.add(*, force=True)."
                )

    def _normalize_value(
        self,
        name: str,
        value: Union[pd.DataFrame, pd.Series],
        force_conversion_to_categorical: bool = False,
    ) -> pd.DataFrame:
        if force_conversion_to_categorical:
            # Convert to Categorical
            warnings.warn(f"Converting {name} annotation to categorical type...")
            return value.astype("category")
        return value


class MetricAttributes(Attributes):
    def __init__(self, mode: Mode, axis: Axis, is_proxy: bool = False, **kwargs):
        """"""
        super().__init__(
            mode=mode,
            axis=axis,
            is_proxy=is_proxy,
            attr_type=AttributeType.METRIC,
            **kwargs,
        )

    def _validate_value(self, value: Union[pd.DataFrame, pd.Series], **kwargs):
        if __DEBUG__:
            print(f"DEBUG: _validate_value ({type(self).__name__})")
        super()._validate_value(value=value)
        _force_conversion_to_numeric = (
            kwargs["force_conversion_to_numeric"]
            if "force_conversion_to_numeric" in kwargs
            else False
        )
        # Do some checks and processing for attribute of type METRIC
        if not _force_conversion_to_numeric:
            if (
                isinstance(value, pd.DataFrame)
                and not all(value.apply(pd.api.types.is_numeric_dtype))
            ) or (
                isinstance(value, pd.Series)
                and not pd.api.types.is_numeric_dtype(arr_or_dtype=value)
            ):
                _dtype = (
                    value.infer_objects().dtypes[0]
                    if isinstance(value, pd.DataFrame)
                    else value.infer_objects().dtype
                )
                raise BadDTypeException(
                    f"Expects value to be numeric but its dtype is {_dtype}"
                )

    def _normalize_value(
        self,
        name: str,
        value: Union[pd.DataFrame, pd.Series],
        force_conversion_to_numeric: bool = False,
    ) -> Union[pd.DataFrame, pd.Series]:
        if force_conversion_to_numeric:
            # Convert to metric
            warnings.warn(f"Converting {name} metric to numeric type...")
            return pd.to_numeric(value)
        return value


class EmbeddingAttributes(Attributes):
    def __init__(self, mode: Mode, axis: Axis, is_proxy: bool = False, **kwargs):
        """"""
        super().__init__(
            mode=mode,
            axis=axis,
            is_proxy=is_proxy,
            attr_type=AttributeType.EMBEDDING,
            **kwargs,
        )


class ClusteringAttributes(Attributes):
    def __init__(self, mode: Mode, axis: Axis, is_proxy: bool = False, **kwargs):
        """"""
        super().__init__(
            mode=mode,
            axis=axis,
            is_proxy=is_proxy,
            attr_type=AttributeType.CLUSTERING,
            **kwargs,
        )


##########################################
# AXIS ATTRIBUTES - FEATURES             #
##########################################


class FeatureAttributes(Attributes):
    def __init__(self, mode: Mode, is_proxy: bool = False):
        """"""
        super().__init__(mode=mode, axis=Axis.FEATURES, is_proxy=is_proxy)

    def _validate_key(self, key: str):
        super()._validate_key(key=key)

    def _validate_value(self, value: Union[pd.DataFrame, pd.Series], **kwargs):
        if __DEBUG__:
            print(f"DEBUG: _validate_value ({type(self).__name__})")
        # Generic validation
        super()._validate_value(value=value)
        # Check if all observations from the given value are present in the DataMatrix of this mode
        _features = self._mode.X._feature_names
        if not all(np.in1d(value.index.astype(str), _features.astype(str))):
            raise Exception(
                f"Cannot add attribute of type {type(value).__name__} to {type(self).__name__}. Index of the given pandas.core.frame.DataFrame does not fully match the features in DataMatrix of mode."
            )

    def __setitem__(self, name, value):
        """"""
        self._validate_key(key=name)
        self._validate_value(value=value)
        # TODO: Attribute type should be inferred here
        _attr = Attribute(
            key=name,
            attr_type=AttributeType.ANNOTATION,
            mode_type=self._mode_type,
            axis=self._axis,
            data=value,
        )
        super()._add_item(key=name, value=_attr)

    @property
    def annotations(self):
        return self._mode._fa_annotations

    def add_annotation(self, key: str, value: Union[pd.DataFrame, pd.Series]):
        self._mode._fa_annotations.add(key=key, value=value)

    @property
    def metrics(self):
        return self._mode._fa_metrics

    def add_metric(self, key: str, value: Union[pd.DataFrame, pd.Series]):
        self._mode._fa_metrics.add(key=key, value=value)


class FeatureAnnotationAttributes(FeatureAttributes, AnnotationAttributes):
    def __init__(self, mode: Mode):
        """"""
        super().__init__(mode=mode, is_proxy=True)

    def __setitem__(self, name, value):
        """"""
        self.add(key=name, value=value)

    def add(self, key: str, value: Union[pd.DataFrame, pd.Series]):
        """"""
        super()._validate_key(key=key)
        super()._validate_value(value=value)
        _data = super()._normalize_value(name=key, value=value)

        _attr = Attribute(
            key=key,
            mode_type=self._mode_type,
            attr_type=self._attr_type,
            axis=self._axis,
            data=_data,
        )

        self._mode._feature_attrs._add_item(key=key, value=_attr)
        super()._add_item_by_value(value=_attr)


class FeatureMetricAttributes(FeatureAttributes, MetricAttributes):
    def __init__(self, mode):
        """"""
        super().__init__(mode=mode, is_proxy=True)

    def __setitem__(self, name: str, value: Union[pd.DataFrame, pd.Series]):
        """"""
        self.add(key=name, value=value)

    def add(self, key: str, value: Union[pd.DataFrame, pd.Series]):
        """"""
        super()._validate_key(key=key)
        super()._validate_value(value=value)
        _data = super()._normalize_value(name=key, value=value)

        _attr = Attribute(
            key=key,
            mode_type=self._mode_type,
            attr_type=self._attr_type,
            axis=self._axis,
            data=_data,
        )

        self._mode._feature_attrs._add_item(key=key, value=_attr)
        super()._add_item_by_value(value=_attr)


##########################################
# AXIS ATTRIBUTES - OBSERVATIONS         #
##########################################


class ObservationAttributes(Attributes):
    def __init__(self, mode: Mode, is_proxy: bool = False, **kwargs):
        """"""
        super().__init__(mode=mode, axis=Axis.OBSERVATIONS, is_proxy=is_proxy, **kwargs)

    def _validate_value(self, value, **kwargs):
        super()._validate_value(value=value, **kwargs)

        # Check if all observations from the given value are present in the DataMatrix of this mode
        _observations = self._mode.X._observation_names
        if not all(np.in1d(value.index, _observations)):
            raise Exception(
                f"Cannot add attribute of type {type(value).__name__} to {type(self).__name__}. Index of the given pandas.core.frame.DataFrame does not fully match the observations in DataMatrix of mode."
            )

    def __setitem__(self, name, value):
        """"""
        super()._validate_key(key=name)
        self._validate_value(value=value)

        # TODO: Attribute type should be inferred here
        _attr = Attribute(
            key=name,
            attr_type=AttributeType.ANNOTATION,
            mode_type=self._mode_type,
            axis=self._axis,
            data=value,
        )
        super()._add_item(key=name, value=_attr)

    @property
    def annotations(self):
        return self._mode._oa_annotations

    def add_annotation(
        self,
        key: str,
        value: Union[pd.DataFrame, pd.Series],
        name: str = None,
        description: str = None,
        force: bool = False,
    ) -> None:
        self._mode._oa_annotations.add(
            key=key, value=value, name=name, description=description, force=force
        )

    @property
    def metrics(self):
        return self._mode._oa_metrics

    def add_metric(
        self,
        key: str,
        value: Union[pd.DataFrame, pd.Series],
        name: str = None,
        description: str = None,
        force: bool = False,
    ) -> None:
        self._mode._oa_metrics.add(
            key=key, value=value, name=name, description=description, force=force
        )

    @property
    def embeddings(self):
        return self._mode._oa_embeddings

    def add_embedding(
        self,
        key: str,
        value: Union[pd.DataFrame, pd.Series],
        name: str = None,
        description: str = None,
    ) -> None:
        self._mode._oa_embeddings.add(
            key=key,
            value=value,
            name=name,
            description=description,
        )

    @property
    def clusterings(self):
        return self._mode._oa_clusterings

    def add_clustering(
        self,
        key: str,
        value: Union[pd.DataFrame, pd.Series],
        name: str = None,
        description: str = None,
    ):
        self._mode._oa_clusterings.add(
            key=key, value=value, name=name, description=description
        )


class ObservationAnnotationAttributes(ObservationAttributes, AnnotationAttributes):
    def __init__(self, mode):
        """"""
        super().__init__(mode=mode, is_proxy=True)

    def __setitem__(self, name, value):
        """"""
        self.add(key=name, value=value)

    def add(
        self, key, value, name: str = None, description: str = None, force: bool = False
    ):
        """"""
        super()._validate_key(key=key)
        super()._validate_value(value=value, force_conversion_to_categorical=force)
        _data = super()._normalize_value(
            name=key, value=value, force_conversion_to_categorical=force
        )

        _attr = Attribute(
            key=key,
            mode_type=self._mode_type,
            attr_type=self._attr_type,
            axis=self._axis,
            data=_data,
            name=name,
            description=description,
        )

        self._mode._observation_attrs._add_item(key=key, value=_attr)
        super()._add_item_by_value(value=_attr)


class ObservationMetricAttributes(ObservationAttributes, MetricAttributes):
    def __init__(self, mode):
        """"""
        super().__init__(mode=mode, is_proxy=True)

    def __setitem__(self, name: str, value: Union[pd.DataFrame, pd.Series]):
        """"""
        self.add(key=name, value=value)

    def add(
        self, key, value, name: str = None, description: str = None, force: bool = False
    ):
        """"""
        super()._validate_key(key=key)
        super()._validate_value(value=value)
        _data = super()._normalize_value(
            name=key, value=value, force_conversion_to_numeric=force
        )

        _attr = Attribute(
            key=key,
            mode_type=self._mode_type,
            attr_type=self._attr_type,
            axis=self._axis,
            data=_data,
            name=name,
            description=description,
        )

        self._mode._observation_attrs._add_item(key=key, value=_attr)
        super()._add_item_by_value(value=_attr)


class EmbeddingAttribute(Attribute):
    def __init__(
        self,
        metadata: LoomXMetadataEmbedding,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._metadata = metadata

    @property
    def id(self) -> int:
        return self._metadata._id

    @id.setter
    def id(self, value: int) -> None:
        self._metadata._id = value

    @property
    def default(self) -> bool:
        return self._metadata.default

    @default.setter
    def default(self, value: bool) -> None:
        self._metadata.default = value

    @property
    def projection_method(self) -> str:
        return self._metadata.projection_metod

    @projection_method.setter
    def projection_method(self, value: str):
        self._metadata.projection_metod = value

    def __repr__(self):
        try:
            _projection_method = ProjectionMethod(self._metadata.projection_method).name
        except:
            _projection_method = "n.a."
        return f"""
{super().__repr__()}
default: {self._metadata.default}
projection method: {_projection_method}
        """


class ObservationEmbeddingAttributes(ObservationAttributes, EmbeddingAttributes):
    def __init__(self, mode):
        """"""
        super().__init__(mode=mode, is_proxy=True, is_multi=True)

    def __setitem__(self, name: str, value: Union[pd.DataFrame, pd.Series]):
        """"""
        self.add(key=name, value=value)

    def add(
        self,
        key: str,
        value: Union[pd.DataFrame, pd.Series],
        name: str = None,
        description: str = None,
        metadata: LoomXMetadataEmbedding = None,
    ):
        super()._validate_key(key=key)
        super()._validate_value(value=value)

        if metadata is None:
            metadata = self.make_metadata(key=key, value=value, name=name)

        _attr = EmbeddingAttribute(
            key=key,
            mode_type=self._mode_type,
            attr_type=self._attr_type,
            axis=self._axis,
            data=value,
            name=name,
            description=description,
            metadata=metadata,
        )
        self._mode._observation_attrs._add_item(key=key, value=_attr)
        super()._add_item_by_value(value=_attr)

    def make_metadata(
        self, key: str, value: Union[pd.DataFrame, pd.Series], name: str = None
    ):
        _embedding_id = len(
            list(
                filter(
                    lambda a: a[1].attr_type == AttributeType.EMBEDDING
                    and int(a[1].id) > 0,
                    self._mode.o,
                )
            )
        )
        return LoomXMetadataEmbedding.from_dict(
            {"id": _embedding_id, "name": key if name is None else name}
        )


class ClusteringAttribute(Attribute):
    def __init__(
        self,
        metadata: LoomXMetadataClustering = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._metadata = metadata

    def __iter__(self):
        """"""
        return iter(ClusteringAttributeIterator(self))

    def __len__(self):
        """"""
        return len(self._keys)

    def __getattribute__(self, name):
        return super().__getattribute__(name)

    def __repr__(self):
        return f"""
{super().__repr__()}
number of clusters: {len(self._metadata.clusters)}
        """

    @property
    def id(self):
        return self._metadata.id

    @id.setter
    def id(self, value: int):
        self._metadata.id = value

    @property
    def group(self):
        return self._metadata.group

    @group.setter
    def group(self, value: str):
        self._metadata.group = value

    @property
    def clusters(self):
        return self._metadata.clusters

    @clusters.setter
    def clusters(self, value: List[LoomXMetadataCluster]):
        self._metadata.clusters = value

    @property
    def clusterMarkerMetrics(self):
        return self._metadata.clusterMarkerMetrics

    @property
    def markers(self):
        return self._metadata.markers

    @property
    def metadata(self) -> LoomXMetadataClustering:
        return self._metadata


class ClusteringAttributeIterator:

    """Class to implement an iterator of ClusteringAttribute """

    def __init__(self, attr: ClusteringAttribute):
        self._attr = attr

    def __iter__(self):
        self._n = 0
        return self

    def __next__(self):
        if self._n < len(self._attr.clusters):
            current_key = self._n
            self._n += 1
            return current_key, self._attr.clusters[current_key]
        else:
            raise StopIteration


class ObservationClusteringAttributes(ObservationAttributes, ClusteringAttributes):
    def __init__(self, mode):
        """"""
        super().__init__(mode=mode, is_proxy=True)

    def __setitem__(self, name: str, value: Union[pd.DataFrame, pd.Series]):
        """"""
        self.add(key=name, value=value)

    def add(
        self,
        key: str,
        value: Union[pd.DataFrame, pd.Series],
        name: str = None,
        description: str = None,
        metadata: LoomXMetadataClustering = None,
    ):
        super()._validate_key(key=key)
        super()._validate_value(value=value)

        _attr = ClusteringAttribute(
            key=key,
            mode_type=self._mode_type,
            attr_type=self._attr_type,
            axis=self._axis,
            data=value,
            name=key if name is None else name,
            description=description,
            metadata=self.make_metadata(key=key, value=value, name=name)
            if metadata is None
            else metadata,
        )
        self._mode._observation_attrs._add_item(key=key, value=_attr)
        super()._add_item_by_value(value=_attr)

    def make_metadata(self, key: str, value: Union[pd.DataFrame, pd.Series], name: str):
        _clusters = []
        for cluster_id in sorted(
            np.unique(value.values).astype(int),
            reverse=False,
        ):
            _clusters.append(LoomXMetadataCluster.from_dict({"id": int(cluster_id)}))
        _clustering_id = len(
            list(
                filter(
                    lambda a: a[1].attr_type == AttributeType.CLUSTERING, self._mode.o
                )
            )
        )
        return LoomXMetadataClustering.from_dict(
            {"id": _clustering_id, "name": key, "group": "n.a.", "clusters": _clusters}
        )



from typing import Tuple

from loomxpy._s7 import S7
from loomxpy._mode import Modes


class LoomX(S7):
    def __init__(self):
        self.modes: Modes = Modes()
        self._active_mode: str = None
        # Data Matrix
        self._data_matrix = None
        # Features
        self._feature_attrs = None
        # Observations
        self._observation_attrs = None

    @property
    def active(self) -> str:
        return self._active_mode

    @active.setter
    def active(self, value) -> None:
        self._validate_active_value(value=value)
        self._active_mode = value
        self._data_matrix = self.modes[value].X
        self._feature_attrs = self.modes[value].f
        self._observation_attrs = self.modes[value].o

    def _validate_active_value(self, value):
        _mode_keys_str = f"{', '.join(self.modes._keys)}"
        if len(self.modes._keys) == 0:
            raise Exception(
                "Cannot set an active mode. None has been detected. You can add one using `loomx.modes.<mode-name> = <data-matrix>`."
            )
        if isinstance(value, Tuple[str, ...]):
            raise Exception("This is currently not implemented.")
        if isinstance(value, str) and value in self.modes:
            return True
        raise Exception(
            f"The mode {value} does not exist. Choose one of: {_mode_keys_str}."
        )

    def _check_active_mode_is_set(self):
        if self._active_mode is None:
            raise Exception(
                "No active mode set. Use `loomx.active = <mode-name>` to set one."
            )

    @property
    def X(self):
        self._check_active_mode_is_set()
        return self._data_matrix

    @property
    def f(self):
        self._check_active_mode_is_set()
        return self._feature_attrs

    @property
    def o(self):
        self._check_active_mode_is_set()
        return self._observation_attrs



import abc


class S7Main(abc.ABC):
    @property
    @abc.abstractproperty
    def X(self):
        raise NotImplementedError()


class S7SideCars(abc.ABC):
    @property
    @abc.abstractproperty
    def f(self):
        raise NotImplementedError()

    @property
    @abc.abstractproperty
    def o(self):
        raise NotImplementedError()


class S7(S7Main, S7SideCars):
    pass



import zlib
import base64
import json
import pandas as pd
import numpy as np


def df_to_named_matrix(df: pd.DataFrame):
    # Create meta-data structure.
    # Create a numpy structured array
    return np.array(
        [tuple(row) for row in df.values],
        dtype=np.dtype(list(zip(df.columns, df.dtypes))),
    )


def compress_encode(value):
    """
    Compress using ZLIB algorithm and encode the given value in base64.
    Taken from: https://github.com/aertslab/SCopeLoomPy/blob/5438da52c4bcf48f483a1cf378b1eaa788adefcb/src/scopeloompy/utils/__init__.py#L7
    """
    return base64.b64encode(zlib.compress(value.encode("ascii"))).decode("ascii")


def decompress_decode(value):
    try:
        value = value.decode("ascii")
        return json.loads(zlib.decompress(base64.b64decode(value)))
    except AttributeError:
        return json.loads(
            zlib.decompress(base64.b64decode(value.encode("ascii"))).decode("ascii")
        )



from abc import ABCMeta


class WithInitHook(ABCMeta):
    def __call__(cls, *args, **kwargs):
        instance = super().__call__(*args, **kwargs)
        instance._initialized = True
        return instance



#!/usr/bin/env python3

import os
import numpy as np
import pandas as pd
import loompy as lp
from sklearn.manifold import TSNE
from pyscenic.aucell import aucell
from pyscenic.genesig import Regulon, GeneSignature
from typing import List, Mapping, Sequence, Optional, Dict
from operator import attrgetter
from multiprocessing import cpu_count
from pyscenic.binarization import binarize
from itertools import chain, repeat, islice
import networkx as nx
import zlib
import base64
import json
import re
import sys

from collections import OrderedDict


class Embedding:
    def __init__(
        self, embedding: pd.DataFrame, embedding_name: str, is_default: bool = False
    ):
        self.embedding = embedding
        self.embedding_name = embedding_name
        self._is_default = is_default

    def get_embedding_name(self):
        return self.embedding_name

    def get_embedding(self):
        return self.embedding

    def is_default(self):
        return self._is_default


class SCopeLoom:
    def __init__(
        self,
        filename: str = None,
        out_fname: str = None,
        ex_mtx: pd.DataFrame = None,
        regulons: List[Regulon] = None,
        cell_annotations: Optional[Mapping[str, str]] = None,
        tree_structure: Sequence[str] = None,
        title: Optional[str] = None,
        nomenclature: str = "Unknown",
        num_workers: int = cpu_count(),
        auc_mtx=None,
        auc_regulon_weights_key="gene2weight",
        auc_thresholds=None,
        compress: bool = False,
        save_additional_regulon_meta_data: bool = False,  # Should be set to true only for multi-runs SCENIC
        set_generic_loom: bool = False,
        tag: str = None,  # Used when merging track and motif-based SCENIC run
        col_attrs: Mapping[str, np.ndarray] = None,
        row_attrs: Mapping[str, np.ndarray] = None,
        global_attrs: Dict = None,
        embeddings: Mapping[str, pd.DataFrame] = None,
    ):
        self.filename = filename
        self.out_fname = out_fname
        self.ex_mtx = ex_mtx
        self.regulons = regulons
        self.cell_annotations = cell_annotations
        self.tree_structure = tree_structure if tree_structure else ()
        self.title = title
        self.nomenclature = nomenclature
        self.num_workers = num_workers
        self.auc_mtx = auc_mtx
        self.auc_regulon_weights_key = auc_regulon_weights_key
        self.auc_thresholds = auc_thresholds
        self.compress = compress
        self.save_additional_regulon_meta_data = save_additional_regulon_meta_data
        self.tag = tag
        self.regulon_filter = None
        # Loom representation
        self.col_attrs = col_attrs if col_attrs else {}
        self.row_attrs = row_attrs if row_attrs else {}
        self.global_attrs = global_attrs if global_attrs else {}
        # Internal representation
        self.embeddings = embeddings if embeddings else {}
        # Utils
        self.id2name = OrderedDict()
        # Set Base Loom
        if set_generic_loom:
            self.set_generic_loom()

    def set_generic_loom(self):
        if self.regulons[0].name.find("_") == -1:
            print(
                "Regulon name does not seem to be compatible with SCOPE. It should include a space to allow selection of the TF.",
                "\nPlease run: \n regulons = [r.rename(r.name.replace('(+)','_('+str(len(r))+'g)')) for r in regulons]",
                "\nor:\n regulons = [r.rename(r.name.replace('(','_(')) for r in regulons]",
            )
        self.set_cell_annotations()
        if self.auc_mtx is None:
            self.auc_mtx = self.calculate_regulon_enrichment()
        if self.auc_thresholds is None:
            self.auc_thresholds = self.binarize_regulon_enrichment()
        if len(self.embeddings.keys()) == 0:
            self.create_loom_default_embedding()
        self.regulon_gene_assignment = self.create_loom_regulon_gene_assignment()
        self.ngenes = self.calculate_nb_genes_per_cell()
        self.clusterings = self.create_loom_clusterings()
        self.regulon_thresholds = self.create_loom_md_regulon_thresholds()
        self.set_generic_col_attrs()
        self.set_generic_row_attrs()
        self.set_generic_global_attrs()
        self.set_tree()

    #######
    # I/O #
    #######

    @staticmethod
    def read_loom(filename: str, tag: str = None):
        with lp.connect(filename, mode="r", validate=False) as loom:

            # Load the content into memory
            # Set the main matrix
            ex_mtx = pd.DataFrame(
                loom[:, :], index=loom.ra.Gene, columns=loom.ca.CellID
            ).T
            # Set the column, row and global attribute using the underlying Dict of the AttributeManager
            col_attrs = {k: v for k, v in loom.ca.items()}
            row_attrs = {k: v for k, v in loom.ra.items()}
            global_attrs = {k: v for k, v in loom.attrs.items()}
            # Decompress and decode the MetaData global attribute
            try:
                global_attrs["MetaData"] = SCopeLoom.decompress_decode(
                    value=global_attrs["MetaData"]
                )
            except Exception:
                # MetaData is uncompressed
                global_attrs["MetaData"] = json.loads(global_attrs["MetaData"])

        scope_loom = SCopeLoom(
            filename=filename,
            ex_mtx=ex_mtx,
            col_attrs=col_attrs,
            row_attrs=row_attrs,
            global_attrs=global_attrs,
            tag=tag,
        )
        if "embeddings" in scope_loom.get_meta_data():
            scope_loom.convert_loom_embeddings_repr_to_internal_repr()

        # If multi-runs mode
        is_multi_runs_mode = scope_loom.has_scenic_multi_runs_data()
        if is_multi_runs_mode:
            scope_loom.set_scenic_min_genes_regulon(
                min_genes_regulon=global_attrs["MetaData"]["regulonSettings"][
                    "min_genes_regulon"
                ]
            )
            scope_loom.set_scenic_min_regulon_gene_occurrence(
                min_regulon_gene_occurrence=global_attrs["MetaData"]["regulonSettings"][
                    "min_regulon_gene_occurrence"
                ]
            )

        return scope_loom

    #############
    # Meta Data #
    #############

    def add_meta_data(self, _dict):
        md = self.global_attrs["MetaData"]
        md.update(_dict)
        self.global_attrs["MetaData"] = md

    def get_meta_data(self):
        return self.global_attrs["MetaData"]

    def set_tree(self):
        assert len(self.tree_structure) <= 3, ""
        self.global_attrs.update(
            ("SCopeTreeL{}".format(idx + 1), category)
            for idx, category in enumerate(
                list(islice(chain(self.tree_structure, repeat("")), 3))
            )
        )

    #############
    # Features  #
    #############

    def get_genes(self):
        return self.row_attrs["Gene"]

    ################
    # Observations #
    ################

    def get_cell_ids(self):
        return self.col_attrs["CellID"]

    ###############
    # Annotations #
    ###############

    def set_cell_annotations(self):
        if self.cell_annotations is None:
            self.cell_annotations = dict(
                zip(self.ex_mtx.index.astype(str), ["-"] * self.ex_mtx.shape[0])
            )

    ###########
    # Metrics #
    ###########

    def calculate_nb_genes_per_cell(self):
        # Calculate the number of genes per cell.
        binary_mtx = self.ex_mtx.copy()
        binary_mtx[binary_mtx != 0] = 1.0
        return binary_mtx.sum(axis=1).astype(int)

    def add_metrics(self, metrics: List[str]):
        md_metrics = []
        for metric in metrics:
            md_metrics.append({"name": metric})
        self.global_attrs["MetaData"].update({"metrics": md_metrics})

    ##############
    # Embeddings #
    ##############

    @staticmethod
    def get_embedding_id(embedding: Embedding, _list):
        """Returns the appropriate index as a string given the _list

        Parameters:
            None

        Returns:
            str: Returns -1 if the given embedding is the default, 0 if the given _list is empty and length of the given _list minus 1

        """
        if embedding.is_default():
            return "-1"
        elif len(_list) == "0":
            return "0"
        else:
            return str(len(_list) - 1)

    def convert_loom_embeddings_repr_to_internal_repr(self):
        for embedding in self.get_meta_data()["embeddings"]:
            self.add_embedding(
                embedding=self.get_embedding_by_id(embedding_id=embedding["id"]),
                embedding_name=embedding["name"],
                is_default=True if str(embedding["id"]) == "-1" else False,
            )

    def get_embedding_by_id(self, embedding_id):
        if str(embedding_id) == "-1":
            return self.col_attrs["Embedding"]
        x = self.col_attrs["Embeddings_X"][str(embedding_id)]
        y = self.col_attrs["Embeddings_Y"][str(embedding_id)]
        return np.column_stack((x, y))

    def has_embedding(self, embedding_name):
        return embedding_name in self.embeddings.keys()

    def add_embedding(
        self, embedding: np.ndarray, embedding_name, is_default: bool = False
    ):
        df_embedding = pd.DataFrame(
            embedding, columns=["_X", "_Y"], index=self.ex_mtx.index
        )
        _embedding = Embedding(
            embedding=df_embedding, embedding_name=embedding_name, is_default=is_default
        )
        if is_default:
            self.default_embedding = _embedding
        self.embeddings[embedding_name] = _embedding

    def create_loom_default_embedding(self):
        # Create an embedding based on tSNE.
        # Name of columns should be "_X" and "_Y".
        self.add_embedding(
            embedding=TSNE().fit_transform(self.auc_mtx),
            embedding_name="tSNE (default)",
            is_default=True,
        )

    def create_loom_md_embeddings_repr(self):
        """Returns a Dictionary (Dict) for the global meta data embeddings with preformated data to be stored in Loom file format and compatible with SCope.

        Parameters:
            None

        Returns:
            dict: A Dictionary (Dict) for the global meta data embeddings with preformated data to be stored in Loom file format and compatible with SCope.

        """
        md_embeddings = []
        for _, embedding in self.embeddings.items():
            if embedding.is_default():
                md_embeddings = md_embeddings + [
                    {"id": "-1", "name": embedding.get_embedding_name()}
                ]
            else:
                md_embeddings = md_embeddings + [
                    {
                        "id": SCopeLoom.get_embedding_id(
                            embedding=embedding, _list=md_embeddings
                        ),
                        "name": embedding.get_embedding_name(),
                    }
                ]
        return {"embeddings": md_embeddings}

    def create_loom_ca_embeddings_repr(self):
        """Returns a Dictionary (Dict) for the embeddings with preformated data to be stored in Loom file format and compatible with SCope.

        Parameters:
            None

        Returns:
            dict: A Dictionary (Dict) for the embeddings with preformated data to be stored in Loom file format and compatible with SCope.

        """
        default_embedding = None
        embeddings_X = pd.DataFrame(index=self.ex_mtx.index)
        embeddings_Y = pd.DataFrame(index=self.ex_mtx.index)

        for _, embedding in self.embeddings.items():
            if embedding.get_embedding().shape[1] != 2:
                raise Exception("The embedding should have two columns.")
            # Set the default embedding
            if embedding.is_default():
                default_embedding = embedding.get_embedding()
            # Update the Embeddings_[X|Y]
            embedding_id = str(
                SCopeLoom.get_embedding_id(embedding, embeddings_X.columns)
            )
            embedding = embedding.get_embedding().copy()
            embedding.columns = ["_X", "_Y"]
            embeddings_X = pd.merge(
                embeddings_X,
                embedding["_X"]
                .to_frame()
                .rename(columns={"_X": embedding_id})
                .astype("float32"),
                left_index=True,
                right_index=True,
            )
            embeddings_Y = pd.merge(
                embeddings_Y,
                embedding["_Y"]
                .to_frame()
                .rename(columns={"_Y": embedding_id})
                .astype("float32"),
                left_index=True,
                right_index=True,
            )
        return {
            "Embedding": SCopeLoom.df_to_named_matrix(df=default_embedding),
            "Embeddings_X": SCopeLoom.df_to_named_matrix(df=embeddings_X),
            "Embeddings_Y": SCopeLoom.df_to_named_matrix(df=embeddings_Y),
        }

    ###############
    # Clusterings #
    ###############

    def create_loom_clusterings(self):
        # Encode cell type clusters.
        # The name of the column should match the identifier of the clustering.
        return (
            pd.DataFrame(
                data=self.ex_mtx.index.values, index=self.ex_mtx.index, columns=["0"]
            )
            .replace(self.cell_annotations)
            .replace(self.name2idx())
        )

    ##########
    # SCENIC #
    ##########

    def set_scenic_min_genes_regulon(self, min_genes_regulon):
        self.scenic_min_genes_regulon = min_genes_regulon

    def set_scenic_min_regulon_gene_occurrence(self, min_regulon_gene_occurrence):
        self.scenic_min_regulon_gene_occurrence = min_regulon_gene_occurrence

    def set_regulon_filter(self, regulons):
        self.regulon_filter = regulons

    def has_scenic_multi_runs_data(self):
        return (
            (
                "RegulonGeneOccurrences" in self.row_attrs.keys()
                and "RegulonGeneWeights" in self.row_attrs.keys()
            )
            or (
                "MotifRegulonGeneOccurrences" in self.row_attrs.keys()
                and "MotifRegulonGeneWeights" in self.row_attrs.keys()
            )
            or (
                "TrackRegulonGeneOccurrences" in self.row_attrs.keys()
                and "TrackRegulonGeneWeights" in self.row_attrs.keys()
            )
        )

    def scopify_md_regulon_data(self):
        # Fix regulon objects to display properly in SCope:
        # Rename regulons in the thresholds object, motif
        md_regulon_thresholds = self.global_attrs["MetaData"]["regulonThresholds"]
        for _, x in enumerate(md_regulon_thresholds):
            tmp = re.sub(r"(_?)\(", "_(", x.get("regulon"))  # + '-motif'
            x.update({"regulon": tmp})
        return {"regulonThresholds": md_regulon_thresholds}

    def scopify_loom_ca_regulon_data(self):
        regulons_auc_col_attrs = list(
            filter(
                lambda col_attrs_key: "RegulonsAUC" in col_attrs_key,
                self.col_attrs.keys(),
            )
        )

        def fix(col_attrs_key):
            regulons = pd.DataFrame(
                self.col_attrs[col_attrs_key], index=self.col_attrs["CellID"]
            )
            # Add underscore for SCope compatibility:
            regulons.columns = regulons.columns.str.replace("_?\\(", "_(")
            return {col_attrs_key: SCopeLoom.df_to_named_matrix(regulons)}

        # Update the keys
        regulons_auc_col_attrs_update = map(fix, regulons_auc_col_attrs)
        # Convert list of dict to dict
        return {
            next(iter(x)): x.get(next(iter(x))) for x in regulons_auc_col_attrs_update
        }

    def scopify_loom_ra_regulon_data(self):
        regulons_row_attrs = list(
            filter(
                lambda row_attrs_key: "Regulon" in row_attrs_key, self.row_attrs.keys()
            )
        )

        def fix(row_attrs_key):
            regulons = pd.DataFrame(
                self.row_attrs[row_attrs_key], index=self.row_attrs["Gene"]
            )
            # Add underscore for SCope compatibility:
            regulons.columns = regulons.columns.str.replace("_?\\(", "_(")
            return {row_attrs_key: SCopeLoom.df_to_named_matrix(regulons)}

        # Update the keys
        regulons_row_attrs_update = map(fix, regulons_row_attrs)
        # Convert list of dict to dict
        return {next(iter(x)): x.get(next(iter(x))) for x in regulons_row_attrs_update}

    def calculate_regulon_enrichment(self):
        # Calculate regulon enrichment per cell using AUCell.
        # Create regulons with weight based on given key
        print(
            "Using {} to weight the genes when running AUCell.".format(
                self.auc_regulon_weights_key
            )
        )
        regulon_signatures = list(
            map(
                lambda x: GeneSignature(
                    name=x.name,
                    gene2weight=self.get_regulon_gene_data(
                        x, self.auc_regulon_weights_key
                    ),
                ),
                self.regulons,
            )
        )
        auc_mtx = aucell(
            self.ex_mtx, regulon_signatures, num_workers=self.num_workers
        )  # (n_cells x n_regulons)
        auc_mtx = auc_mtx.loc[self.ex_mtx.index]
        return auc_mtx

    def binarize_regulon_enrichment(self):
        _, auc_thresholds = binarize(self.auc_mtx)
        return auc_thresholds

    @staticmethod
    def format_tag(tag):
        # return f"{tag[0]}{tag[-1]}"
        return tag

    def get_regulon_gene_weights(self):
        if self.regulon_filter is not None:
            return self.row_attrs["RegulonGeneWeights"][self.regulon_filter]
        return self.row_attrs["RegulonGeneWeights"]

    def get_regulon_gene_occurrences(self):
        if self.regulon_filter is not None:
            return self.row_attrs["RegulonGeneOccurrences"][self.regulon_filter]
        return self.row_attrs["RegulonGeneOccurrences"]

    def get_regulons(self):
        if self.regulon_filter is not None:
            return self.row_attrs["Regulons"][self.regulon_filter]
        return self.row_attrs["Regulons"]

    def merge_regulon_data(self, scope_loom):
        # Check if SCENIC has been run in multi-runs mode
        is_multi_runs_mode = (
            self.has_scenic_multi_runs_data()
            and scope_loom.has_scenic_multi_runs_data()
        )

        # RegulonsAUC
        # Relabel columns with suffix indicating the regulon source
        auc_mtx = pd.DataFrame(
            data=self.col_attrs["RegulonsAUC"], index=self.col_attrs["CellID"]
        )
        auc_mtx.columns = auc_mtx.columns + "-" + SCopeLoom.format_tag(self.tag)

        scope_loom_auc_mtx = pd.DataFrame(
            data=scope_loom.col_attrs["RegulonsAUC"],
            index=scope_loom.col_attrs["CellID"],
        )
        scope_loom_auc_mtx.columns = (
            scope_loom_auc_mtx.columns + "-" + SCopeLoom.format_tag(scope_loom.tag)
        )

        # Regulons (regulon assignment matrices)
        regulons = pd.DataFrame(self.get_regulons(), index=self.row_attrs["Gene"])
        regulons.columns = regulons.columns + "-" + SCopeLoom.format_tag(self.tag)

        scope_loom_regulons = pd.DataFrame(
            scope_loom.get_regulons(), index=scope_loom.row_attrs["Gene"]
        )
        scope_loom_regulons.columns = (
            scope_loom_regulons.columns + "-" + SCopeLoom.format_tag(scope_loom.tag)
        )

        # If multi-runs SCENIC
        # Rename Regulons Gene Occurrences
        if is_multi_runs_mode:
            regulon_gene_occurrences = pd.DataFrame(
                self.get_regulon_gene_occurrences(), index=self.row_attrs["Gene"]
            )
            regulon_gene_occurrences.columns = (
                regulon_gene_occurrences.columns + "-" + SCopeLoom.format_tag(self.tag)
            )

            scope_loom_regulon_gene_occurrences = pd.DataFrame(
                scope_loom.get_regulon_gene_occurrences(),
                index=scope_loom.row_attrs["Gene"],
            )
            scope_loom_regulon_gene_occurrences.columns = (
                scope_loom_regulon_gene_occurrences.columns
                + "-"
                + SCopeLoom.format_tag(scope_loom.tag)
            )

        # If multi-runs SCENIC
        # Rename Regulons Gene Weights
        if is_multi_runs_mode:
            regulon_gene_weights = pd.DataFrame(
                self.get_regulon_gene_weights(), index=self.row_attrs["Gene"]
            )
            regulon_gene_weights.columns = (
                regulon_gene_weights.columns + "-" + SCopeLoom.format_tag(self.tag)
            )

            scope_loom_regulon_gene_weights = pd.DataFrame(
                scope_loom.get_regulon_gene_weights(),
                index=scope_loom.row_attrs["Gene"],
            )
            scope_loom_regulon_gene_weights.columns = (
                scope_loom_regulon_gene_weights.columns
                + "-"
                + SCopeLoom.format_tag(scope_loom.tag)
            )

        # Combine meta data regulons
        # Rename regulons in the thresholds object, motif
        rt = self.global_attrs["MetaData"]["regulonThresholds"]
        for _, x in enumerate(rt):
            tmp = x.get("regulon") + "-" + SCopeLoom.format_tag(self.tag)
            x.update({"regulon": tmp})

        # Rename regulons in the thresholds object, track
        scope_loom_rt = scope_loom.global_attrs["MetaData"]["regulonThresholds"]
        for _, x in enumerate(scope_loom_rt):
            tmp = x.get("regulon") + "-" + SCopeLoom.format_tag(scope_loom.tag)
            x.update({"regulon": tmp})
            # blank out the "motifData" field for track-based regulons:
            x.update({"mofitData": "NA.png"})

        # merge regulon threshold dictionaries:
        rt_merged = rt + scope_loom_rt

        # Remove because we will save them separately
        del self.row_attrs["Regulons"]
        del self.col_attrs["RegulonsAUC"]

        if is_multi_runs_mode:
            del self.row_attrs["RegulonGeneOccurrences"]
            del self.row_attrs["RegulonGeneWeights"]

        # Update the attributes
        self.row_attrs.update(
            {f"{self.tag.capitalize()}Regulons": SCopeLoom.df_to_named_matrix(regulons)}
        )
        self.col_attrs.update(
            {
                f"{self.tag.capitalize()}RegulonsAUC": SCopeLoom.df_to_named_matrix(
                    auc_mtx
                )
            }
        )

        if is_multi_runs_mode:
            self.row_attrs.update(
                {
                    f"{self.tag.capitalize()}RegulonGeneOccurrences": SCopeLoom.df_to_named_matrix(
                        regulon_gene_occurrences
                    )
                }
            )
            self.row_attrs.update(
                {
                    f"{self.tag.capitalize()}RegulonGeneWeights": SCopeLoom.df_to_named_matrix(
                        regulon_gene_weights
                    )
                }
            )

        self.row_attrs.update(
            {
                f"{scope_loom.tag.capitalize()}Regulons": SCopeLoom.df_to_named_matrix(
                    scope_loom_regulons
                )
            }
        )
        self.col_attrs.update(
            {
                f"{scope_loom.tag.capitalize()}RegulonsAUC": SCopeLoom.df_to_named_matrix(
                    scope_loom_auc_mtx
                )
            }
        )

        if is_multi_runs_mode:
            self.row_attrs.update(
                {
                    f"{scope_loom.tag.capitalize()}RegulonGeneOccurrences": SCopeLoom.df_to_named_matrix(
                        scope_loom_regulon_gene_occurrences
                    )
                }
            )
            self.row_attrs.update(
                {
                    f"{scope_loom.tag.capitalize()}RegulonGeneWeights": SCopeLoom.df_to_named_matrix(
                        scope_loom_regulon_gene_weights
                    )
                }
            )

        self.global_attrs["MetaData"].update({"regulonThresholds": rt_merged})

    def get_regulon_gene_data(self, regulon, key):
        if key == "gene2occurrence":
            return regulon.gene2occurrence
        if key == "gene2weight":
            return regulon.gene2weight
        raise Exception(
            "Cannot retrieve {} from given regulon. Not implemented.".format(key)
        )

    def get_regulon_meta_data(self, name, threshold):
        def fetch_logo(context):
            for elem in context:
                if elem.endswith(".png"):
                    return elem
            return ""

        name2logo = {reg.name: fetch_logo(reg.context) for reg in self.regulons}

        regulon = list(filter(lambda x: x.name == name, self.regulons))[0]
        regulon_meta_data = {
            "regulon": name,
            "defaultThresholdValue": (
                threshold if isinstance(threshold, float) else threshold[0]
            ),
            "defaultThresholdName": "guassian_mixture_split",
            "allThresholds": {
                "guassian_mixture_split": (
                    threshold if isinstance(threshold, float) else threshold[0]
                )
            },
            "motifData": name2logo.get(name, ""),
        }
        if self.save_additional_regulon_meta_data:
            regulon_meta_data.update(
                {
                    "tf": regulon.transcription_factor,
                    "score": regulon.score if hasattr(regulon, "score") else 0.0,
                    "orthologousIdentity": regulon.orthologous_identity
                    if hasattr(regulon, "orthologous_identity")
                    else 0.0,
                    "annotation": regulon.annotation
                    if hasattr(regulon, "annotation")
                    else "",
                    "similarityQValue": regulon.score
                    if hasattr(regulon, "similarity_qvalue")
                    else 0.0,
                }
            )
        return regulon_meta_data

    def create_loom_md_regulon_thresholds(self):
        return [
            self.get_regulon_meta_data(name, threshold)
            for name, threshold in self.auc_thresholds.iteritems()
        ]

    def create_loom_regulon_gene_assignment(self):
        # Encode genes in regulons as "binary" membership matrix.
        genes = np.array(self.ex_mtx.columns)
        n_genes = len(genes)
        n_regulons = len(self.regulons)
        data = np.zeros(shape=(n_genes, n_regulons), dtype=int)
        for idx, regulon in enumerate(self.regulons):
            data[:, idx] = np.isin(genes, regulon.genes).astype(int)
        regulon_gene_assignment = pd.DataFrame(
            data=data,
            index=self.ex_mtx.columns,
            columns=list(map(attrgetter("name"), self.regulons)),
        )
        return regulon_gene_assignment

    # Multi-runs SCENIC additional data

    def encode_regulon_gene_data(self, key):
        genes = np.array(self.ex_mtx.columns)
        n_genes = len(genes)
        n_regulons = len(self.regulons)
        data = np.zeros(shape=(n_genes, n_regulons), dtype=float)
        for idx, regulon in enumerate(self.regulons):
            regulon_data = pd.DataFrame.from_dict(
                self.get_regulon_gene_data(regulon, key), orient="index"
            )
            data[np.isin(genes, regulon_data.index), idx] = regulon_data[0][
                genes[np.isin(genes, regulon_data.index)]
            ]
        return data

    def create_loom_regulon_gene_data(self, key):
        return pd.DataFrame(
            data=self.encode_regulon_gene_data(key=key),
            index=self.ex_mtx.columns,
            columns=list(map(attrgetter("name"), self.regulons)),
        )

    def add_row_attr_regulon_gene_weights(self):
        regulon_gene_weights = self.create_loom_regulon_gene_data(key="gene2weight")
        if "RegulonGeneWeights" not in self.row_attrs.keys():
            print("Added 'RegulonGeneWeights' to the row attributes.")
            self.row_attrs["RegulonGeneWeights"] = SCopeLoom.df_to_named_matrix(
                regulon_gene_weights
            )

    def add_row_attr_regulon_gene_occurrences(self):
        regulon_gene_occurrences = self.create_loom_regulon_gene_data(
            key="gene2occurrence"
        )
        if "RegulonGeneOccurrences" not in self.row_attrs.keys():
            print("Added 'RegulonGeneOccurrences' to the row attributes.")
            self.row_attrs["RegulonGeneOccurrences"] = SCopeLoom.df_to_named_matrix(
                regulon_gene_occurrences
            )

    ###########
    # Generic #
    ###########

    def name2idx(self):
        return dict(
            map(reversed, enumerate(sorted(set(self.cell_annotations.values()))))
        )

    def set_generic_col_attrs(self):
        self.col_attrs = {
            "CellID": self.ex_mtx.index.values.astype("str"),
            "nGene": self.ngenes.values,
            "Embedding": SCopeLoom.df_to_named_matrix(
                self.default_embedding.get_embedding()
            ),
            "RegulonsAUC": SCopeLoom.df_to_named_matrix(self.auc_mtx),
            "Clusterings": SCopeLoom.df_to_named_matrix(self.clusterings),
            "ClusterID": self.clusterings.values,
        }

    def set_generic_row_attrs(self):
        self.row_attrs = {
            "Gene": self.ex_mtx.columns.values.astype("str"),
            "Regulons": SCopeLoom.df_to_named_matrix(self.regulon_gene_assignment),
        }

    def set_generic_global_attrs(self):
        self.global_attrs = {
            "title": os.path.splitext(os.path.basename(self.out_fname))[0]
            if self.title is None
            else self.title,
            "MetaData": {
                "embeddings": [
                    {"id": identifier, "name": name}
                    for identifier, name in self.id2name.items()
                ],
                "annotations": [{"name": "", "values": []}],
                "clusterings": [
                    {
                        "id": 0,
                        "group": "celltype",
                        "name": "Cell Type",
                        "clusters": [
                            {"id": idx, "description": name}
                            for name, idx in self.name2idx().items()
                        ],
                    }
                ],
                "regulonThresholds": self.regulon_thresholds,
            },
            "Genome": self.nomenclature,
        }

    def sort(self, axis, by):
        if axis == 0:
            if by not in self.row_attrs:
                sys.exit(f"ERROR: Cannot sort on a row attribute that does not exist")
            sorted_idx = np.argsort(self.row_attrs[by])
            for ra_key in self.row_attrs.keys():
                self.row_attrs[ra_key] = self.row_attrs[ra_key][sorted_idx]
            self.ex_mtx = self.ex_mtx[self.row_attrs[by]]
        else:
            sys.exit(f"ERROR: Sorting by column axis is not implemented")

    def merge(self, loom, overwrite_embeddings=False):
        # Check all the cells and genes are in the same order
        if not np.array_equal(a1=self.get_cell_ids(), a2=loom.get_cell_ids()):
            sys.exit(
                f"ERROR: Column attribute CellIDs does not match between {os.path.basename(self.filename)} and {os.path.basename(loom.filename)}"
            )
        if not np.array_equal(a1=self.get_genes(), a2=loom.get_genes()):
            sys.exit(
                f"ERROR: Row attribute 'Gene' does not match between {os.path.basename(self.filename)} and {os.path.basename(loom.filename)}"
            )
        if not np.array_equal(a1=self.get_genes(), a2=self.ex_mtx.columns):
            sys.exit(
                f"ERROR: Row attribute 'Gene' does not match with the features in the matrix of {os.path.basename(self.filename)}."
            )
        if not np.array_equal(a1=loom.get_genes(), a2=loom.ex_mtx.columns):
            sys.exit(
                f"ERROR: Row attribute 'Gene' does not match with the features in the matrix of {os.path.basename(loom.filename)}."
            )

        # Add the embeddings of the given loom (SCopeLoom) to this SCopeLoom
        if "embeddings" in loom.get_meta_data():

            print(
                f"Merging embeddings from {os.path.basename(loom.filename)} into {os.path.basename(self.filename)}..."
            )
            for embedding in loom.get_meta_data()["embeddings"]:
                # Only add the embedding if not present in this loom and overwrite_embeddings is False
                if (
                    not self.has_embedding(embedding_name=embedding["name"])
                    or overwrite_embeddings
                ):
                    print(f"Adding {embedding['name']} embedding...")
                    self.add_embedding(
                        embedding_name=embedding["name"],
                        embedding=loom.get_embedding_by_id(
                            embedding_id=embedding["id"]
                        ),
                    )
            print("Done.")

        # Add SCENIC results if present in the given loom
        if any("Regulon" in s for s in loom.row_attrs.keys()) and any(
            "Regulon" in s for s in loom.col_attrs.keys()
        ):

            print(
                f"Merging SCENIC results from {os.path.basename(loom.filename)} into {os.path.basename(self.filename)}..."
            )

            # Regulon information from row attributes
            # All the row attributes containing the substring 'Regulon' within the given SCopeLoom
            # will be merged with this SCopeLoom
            regulons_row_attrs_keys = list(
                filter(
                    lambda row_attrs_key: "Regulon" in row_attrs_key,
                    loom.row_attrs.keys(),
                )
            )
            regulons_row_attrs = {
                regulons_row_attr_key: loom.row_attrs[regulons_row_attr_key]
                for regulons_row_attr_key in regulons_row_attrs_keys
            }
            self.row_attrs.update(regulons_row_attrs)
            # Regulon information from column attributes
            # All the column attributes containing the substring 'Regulon' within the given SCopeLoom
            # will be merged with this SCopeLoom
            regulons_auc_col_attrs_keys = list(
                filter(
                    lambda col_attrs_key: "Regulon" in col_attrs_key,
                    loom.col_attrs.keys(),
                )
            )
            regulons_auc_col_attrs = {
                regulons_auc_col_attr_key: loom.col_attrs[regulons_auc_col_attr_key]
                for regulons_auc_col_attr_key in regulons_auc_col_attrs_keys
            }
            self.col_attrs.update(regulons_auc_col_attrs)
            # regulonThresholds MetaData global attribute
            self.global_attrs["MetaData"].update(
                {"regulonThresholds": loom.get_meta_data()["regulonThresholds"]}
            )

            print("Done.")

        # If multi-runs mode
        is_multi_runs_mode = self.has_scenic_multi_runs_data()
        if is_multi_runs_mode:
            self.set_scenic_min_genes_regulon(
                min_genes_regulon=loom.scenic_min_genes_regulon
            )
            self.set_scenic_min_regulon_gene_occurrence(
                min_regulon_gene_occurrence=loom.scenic_min_regulon_gene_occurrence
            )

    def export(
        self,
        out_fname: str,
        save_embeddings: bool = True,
        compress_meta_data: bool = False,
    ):

        if out_fname is None:
            raise ValueError("The given out_fname cannot be None.")

        ##############
        # Embeddings #
        ##############
        if save_embeddings:
            self.col_attrs.update(self.create_loom_ca_embeddings_repr())
            self.global_attrs["MetaData"].update(self.create_loom_md_embeddings_repr())

        ##########
        # SCENIC #
        ##########
        is_multi_runs_mode = self.has_scenic_multi_runs_data()
        if is_multi_runs_mode:
            if "regulonSettings" not in self.global_attrs["MetaData"].keys():
                self.add_meta_data(
                    _dict={
                        "regulonSettings": {
                            "min_genes_regulon": self.scenic_min_genes_regulon,
                            "min_regulon_gene_occurrence": self.scenic_min_regulon_gene_occurrence,
                        }
                    }
                )

        if any("Regulon" in s for s in self.row_attrs.keys()):
            self.row_attrs.update(self.scopify_loom_ra_regulon_data())
        if any("RegulonsAUC" in s for s in self.col_attrs.keys()):
            self.col_attrs.update(self.scopify_loom_ca_regulon_data())
        if "regulonThresholds" in self.global_attrs["MetaData"].keys():
            self.global_attrs["MetaData"].update(self.scopify_md_regulon_data())

        # Compress MetaData global attribute
        # (Should be compressed if Loompy version 2)
        self.global_attrs["MetaData"] = json.dumps(self.global_attrs["MetaData"])
        if compress_meta_data:
            self.global_attrs["MetaData"] = SCopeLoom.compress_encode(
                value=self.global_attrs["MetaData"]
            )

        # Create loom file for use with the SCope tool.
        # The loom file format opted for rows as genes to facilitate growth along the column axis (i.e add more cells)
        # PySCENIC chose a different orientation because of limitation set by the feather format: selectively reading
        # information from disk can only be achieved via column selection. For the ranking databases this is of utmost
        # importance.
        lp.create(
            filename=out_fname,
            layers=self.ex_mtx.T.values,
            row_attrs=self.row_attrs,
            col_attrs=self.col_attrs,
            file_attrs=self.global_attrs,
        )

    #########
    # Utils #
    #########

    @staticmethod
    def df_to_named_matrix(df: pd.DataFrame):
        # Create meta-data structure.
        # Create a numpy structured array
        return np.array(
            [tuple(row) for row in df.values],
            dtype=np.dtype(list(zip(df.columns, df.dtypes))),
        )

    @staticmethod
    def compress_encode(value):
        """
        Compress using ZLIB algorithm and encode the given value in base64.
        Taken from: https://github.com/aertslab/SCopeLoomPy/blob/5438da52c4bcf48f483a1cf378b1eaa788adefcb/src/scopeloompy/utils/__init__.py#L7
        """
        return base64.b64encode(zlib.compress(value.encode("ascii"))).decode("ascii")

    @staticmethod
    def decompress_decode(value):
        try:
            value = value.decode("ascii")
            return json.loads(zlib.decompress(base64.b64decode(value)))
        except AttributeError:
            return json.loads(
                zlib.decompress(base64.b64decode(value.encode("ascii"))).decode("ascii")
            )



from .v1.metadata import (
    GLOBAL_ATTRIBUTE_KEY,
    Metadata as LoomXMetadata,
    Clustering as LoomXMetadataClustering,
    Cluster as LoomXMetadataCluster,
    Embedding as LoomXMetadataEmbedding,
    ProjectionMethod as ProjectionMethod,
    ClusterMarkerMetric as LoomXMetadataClusterMarkerMetric,
)



from enum import Enum
import pandas as pd
from dataclasses_json import dataclass_json, config, DataClassJsonMixin, cfg
from dataclasses import dataclass, field
from typing import Optional, List

GLOBAL_ATTRIBUTE_KEY = "MetaData"


@dataclass_json
@dataclass
class Annotation(DataClassJsonMixin):
    name: str
    values: List[str]


@dataclass_json
@dataclass
class Metric(DataClassJsonMixin):
    name: str


class ProjectionMethod(Enum):
    PCA = 0
    TSNE = 1
    UMAP = 2


@dataclass_json
@dataclass
class Embedding(DataClassJsonMixin):
    _id: int = field(metadata=config(field_name="id"))
    _name: str = field(metadata=config(field_name="name"))
    # Attributes exluded from the metadata
    _default: Optional[bool] = field(
        default=None,
        metadata=config(
            field_name="default", exclude=cfg.Exclude.ALWAYS
        ),  # has to be exluded since not part of SCope gRPC API
    )
    _projection_method: Optional[ProjectionMethod] = field(
        default=None,
        metadata=config(
            field_name="projection_method", exclude=cfg.Exclude.ALWAYS
        ),  # has to be exluded since not part of SCope gRPC API
    )

    def __post_init__(self):
        if self._projection_method is not None:
            if "pca" in self._name.lower():
                self._projection_method = ProjectionMethod.PCA
            elif "tsne" in self._name.lower():
                self._projection_method = ProjectionMethod.TSNE
            elif "umap" in self._name.lower():
                self._projection_method = ProjectionMethod.UMAP

    @property
    def id(self):
        return int(self._id)

    @property
    def default(self):
        if self._default is None:
            return int(self._id) == -1
        return False

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def projection_method(self) -> str:
        if self._projection_method is None:
            return "n.a."
        return self._projection_method.name


@dataclass_json
@dataclass
class RegulonAllThresholds:
    tenPercentOfMax: Optional[float] = None


@dataclass_json
@dataclass
class RegulonThreshold:
    regulon: str
    defaultThresholdValue: float
    defaultThresholdName: str
    allThresholds: RegulonAllThresholds
    motifData: str
    motifVersion: Optional[str] = field(default="")


@dataclass_json
@dataclass
class CellTypeAnnotationVoter:
    voter_name: str
    voter_id: str
    voter_hash: str


@dataclass_json
@dataclass
class CellTypeAnnotationVoters:
    total: int
    voters: List[CellTypeAnnotationVoter]


@dataclass_json
@dataclass
class CellTypeAnnotationVotes:
    votes_for: CellTypeAnnotationVoters
    votes_against: CellTypeAnnotationVoters


@dataclass_json
@dataclass
class CellTypeAnnotationData:
    curator_name: str
    curator_id: str
    timestamp: int
    obo_id: str
    ols_iri: str
    annotation_label: str
    markers: field(default_factory=lambda: [])
    publication: str
    comment: str


@dataclass_json
@dataclass
class CellTypeAnnotation:
    data: CellTypeAnnotationData
    validate_hash: str
    votes: CellTypeAnnotationVotes


@dataclass_json
@dataclass
class ClusterMarkerMetric(DataClassJsonMixin):
    _accessor: str = field(default=None, metadata=config(field_name="accessor"))
    _name: str = field(default=None, metadata=config(field_name="name"))
    _description: str = field(default=None, metadata=config(field_name="description"))
    # Attributes exluded from the metadata
    _threshold: Optional[float] = field(
        default=None,
        metadata=config(field_name="threshold", exclude=cfg.Exclude.ALWAYS),
    )
    _threshold_method: Optional[str] = field(
        default=None,
        metadata=config(field_name="threshold_method", exclude=cfg.Exclude.ALWAYS),
    )

    @property
    def accessor(self) -> str:
        return self._accessor

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> str:
        return self._description


@dataclass_json
@dataclass
class Cluster(DataClassJsonMixin):
    _id: int = field(metadata=config(field_name="id"))
    _description: str = field(default="", metadata=config(field_name="description"))
    _cell_type_annotation: List[CellTypeAnnotation] = field(
        default_factory=lambda: [], metadata=config(field_name="cell_type_annotation")
    )
    # Attributes exluded from the metadata
    _markers: Optional[pd.DataFrame] = field(
        default=None,
        metadata=config(field_name="markers", exclude=cfg.Exclude.ALWAYS),
    )

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, value):
        raise Exception("The ID of the cluster cannot be changed.")

    @property
    def name(self):
        if self._name is not None:
            return self._name
        return f"Unannotated Cluster {self._id}"

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def description(self):
        if self._description is not None:
            return self._description
        return self.name

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def markers(self):
        return self._markers

    @markers.setter
    def markers(self, value: pd.DataFrame):
        self._markers = value


@dataclass_json
@dataclass
class Clustering(DataClassJsonMixin):
    _id: int = field(metadata=config(field_name="id"))
    _group: str = field(metadata=config(field_name="group"))
    _name: str = field(metadata=config(field_name="name"))
    _clusters: List[Cluster] = field(metadata=config(field_name="clusters"))
    _clusterMarkerMetrics: List[ClusterMarkerMetric] = field(
        default_factory=lambda: [], metadata=config(field_name="clusterMarkerMetrics")
    )
    # Attributes exluded from the metadata
    _markers: Optional[pd.DataFrame] = field(
        default=None,
        metadata=config(field_name="markers", exclude=cfg.Exclude.ALWAYS),
    )

    @property
    def id(self) -> int:
        return self._id

    @id.setter
    def id(self, value):
        raise Exception("The ID of the clustering cannot be changed.")

    @property
    def group(self) -> str:
        return self._group

    @group.setter
    def group(self, value: str) -> None:
        self._group = value

    @property
    def name(self):
        if self._name is not None:
            return self._name
        return f"Unannotated Clustering {self._id}"

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def clusters(self) -> List[Cluster]:
        return self._clusters

    @property
    def clusterMarkerMetrics(self) -> List[ClusterMarkerMetric]:
        return self._clusterMarkerMetrics

    @property
    def markers(self):
        if self._markers is None:
            self._markers = pd.concat([cluster.markers for cluster in self._clusters])
        return self._markers


@dataclass_json
@dataclass
class Metadata(DataClassJsonMixin):
    metrics: List[Metric]
    annotations: List[Annotation]
    embeddings: List[Embedding]
    clusterings: Optional[List[Clustering]]
    regulonThresholds: List[RegulonThreshold] = field(default_factory=lambda: [])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['loomxpy',
 'loomxpy._io',
 'loomxpy._specifications',
 'loomxpy._specifications.v1',
 'tests']

package_data = \
{'': ['*']}

install_requires = \
['click',
 'dataclasses-json>=0.5.3,<0.6.0',
 'loompy>=3.0.6,<4.0.0',
 'numpy>=1.20.2,<2.0.0',
 'pandas>=1.2.4,<2.0.0',
 'pyscenic>=0.11.1,<0.12.0',
 'scikit-learn>=0.24.2,<0.25.0']

entry_points = \
{'console_scripts': ['loomxpy = loomxpy.cli:main']}

setup_kwargs = {
    'name': 'loomxpy',
    'version': '0.4.1',
    'description': 'Top-level package for LoomXpy.',
    'long_description': '=======\nLoomXpy\n=======\n\nPython package (compatible with SCope) to create .loom files and extend them with other data e.g.: SCENIC regulons\n\n.. image:: https://img.shields.io/pypi/v/loomxpy.svg\n        :target: https://pypi.python.org/pypi/loomxpy\n\n.. image:: https://readthedocs.org/projects/loomxpy/badge/?version=latest\n        :target: https://loomxpy.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n.. image:: https://pyup.io/repos/github/aertslab/loomxpy/shield.svg\n     :target: https://pyup.io/repos/github/aertslab/loomxpy/\n     :alt: Updates\n\n\n\nPython package containing code for Loom files generation.\n\n\n* Free software: MIT\n* Documentation: https://loomxpy.readthedocs.io.\n\n\nFeatures\n--------\n\n* TODO\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `briggySmalls/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`briggySmalls/cookiecutter-pypackage`: https://github.com/briggySmalls/cookiecutter-pypackage\n',
    'author': 'Maxime De Waegeneer',
    'author_email': 'mdewaegeneer@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aertslab/loomxpy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1',
}


setup(**setup_kwargs)



import numpy as np
from scipy import sparse


class DataMatrix:
    def __init__(self, data_matrix, feature_names, observation_names):
        self._data_matrix = self._normalize_data_matrix(
            data_matrix=data_matrix,
            feature_names=feature_names,
            observation_names=observation_names,
        )
        self._feature_names = feature_names
        self._observation_names = observation_names

    @staticmethod
    def _validate_data_matrix(data_matrix, feature_names, observation_names):
        if feature_names is None:
            raise Exception(
                "The given `data_matrix` argument is a NumPy 2D matrix. Hence, the `feature_names` argument should be provided."
            )
        if observation_names is None:
            raise Exception(
                "The given `data_matrix` argument is a NumPy 2D matrix. Hence, the `observation_names` argument should be provided."
            )
        if (
            len(observation_names) != data_matrix.shape[0]
            or len(feature_names) != data_matrix.shape[1]
        ):
            raise Exception(
                f"""
Expecting the given `data_matrix` argument to have features as rows and observations as columns. Trying to assign {len(observation_names)} observations
and {len(feature_names)} features to data matrix with {data_matrix.shape[0]} rows and {data_matrix.shape[1]} columns.
            """
            )

    @staticmethod
    def _normalize_data_matrix_from_numpy_matrix(
        data_matrix, feature_names, observation_names
    ) -> sparse.csr_matrix:
        DataMatrix._validate_data_matrix(
            data_matrix=data_matrix,
            feature_names=feature_names,
            observation_names=observation_names,
        )
        return sparse.csr_matrix(data_matrix)

    @staticmethod
    def _normalize_data_matrix_from_sparse_matrix(
        data_matrix, feature_names, observation_names
    ) -> sparse.csr_matrix:
        DataMatrix._validate_data_matrix(
            data_matrix=data_matrix,
            feature_names=feature_names,
            observation_names=observation_names,
        )
        return data_matrix

    def _normalize_data_matrix(
        self, data_matrix, feature_names=None, observation_names=None
    ) -> sparse.csr_matrix:
        if isinstance(data_matrix, np.matrix):
            return DataMatrix._normalize_data_matrix_from_numpy_matrix(
                data_matrix=data_matrix,
                feature_names=feature_names,
                observation_names=observation_names,
            )
        if sparse.issparse(data_matrix):
            return DataMatrix._normalize_data_matrix_from_sparse_matrix(
                data_matrix=data_matrix,
                feature_names=feature_names,
                observation_names=observation_names,
            )
        raise Exception("Not a valid `data_matrix` argument.")

    def __repr__(self):
        return f"DataMatrix with {len(self._feature_names)} features and {len(self._observation_names)} observations."



import json
import zlib
import base64
import functools
import pandas as pd
import numpy as np
import loompy as lp
from typing import Dict, Union, Tuple
from scipy import sparse

from loomxpy._errors import BadDTypeException
from loomxpy._loomx import LoomX
from loomxpy._mode import Mode, ModeType
from loomxpy._specifications.v1.metadata import (
    Embedding,
    Metadata as LoomXMetadataV1,
    Clustering as LoomXMetadataV1Clustering,
    Cluster as LoomXMetadataV1Cluster,
    GLOBAL_ATTRIBUTE_KEY as GLOBAL_ATTRIBUTE_KEY_V1,
)

# A list of all versions of the MetaData global attribute
GLOBAL_ATTRIBUTE_KEY_VX = [GLOBAL_ATTRIBUTE_KEY_V1]


def decompress_metadata(metadata: str):
    try:
        metadata = metadata.decode("ascii")
        return json.loads(zlib.decompress(base64.b64decode(s=metadata)))
    except AttributeError:
        return json.loads(
            zlib.decompress(base64.b64decode(metadata.encode("ascii"))).decode("ascii")
        )


def get_metadata_from_loom_connection(
    loom_connection: lp.LoomConnection, metadata_global_attribute_key: str
):
    _metadata = loom_connection.attrs[metadata_global_attribute_key]
    if type(_metadata) is np.ndarray:
        _metadata = _metadata[0]
    try:
        return json.loads(s=_metadata)
    except json.decoder.JSONDecodeError:
        return decompress_metadata(metadata=_metadata)


def _read_scope_metadata(
    loom_connection: lp.LoomConnection,
) -> Union[Tuple["v1", Union[LoomXMetadataV1]]]:
    # Read the Metadata from SCope
    try:
        _metadata: LoomXMetadataV1 = LoomXMetadataV1.from_dict(
            get_metadata_from_loom_connection(
                loom_connection=loom_connection,
                metadata_global_attribute_key=GLOBAL_ATTRIBUTE_KEY_V1,
            )
        )
        return "v1", _metadata
    except:
        pass
    raise Exception("Cannot read the LoomX metadata.")


def _read_scope_v1_rna_loom(
    loom_connection: lp.LoomConnection,
    metadata: LoomXMetadataV1,
    force_conversion: Dict,
):
    # Loompy stores the features as rows and the observations as columns
    # LoomX follows Scanpy convention (i.e.: machine learning/statistics standards) i.e.: observations as rows and the features as columns. This requires to transpose the matrix
    _matrix = loom_connection[:, :]
    if "Gene" not in loom_connection.ra:
        raise Exception("The loom file is missing a 'Gene' row attribute")
    if "CellID" not in loom_connection.ca:
        raise Exception("The loom file is missing a 'CellID' column attribute")
    # Create the LoomX in-memory object
    ## Add the expression matrix in the RNA mode
    _lx = LoomX()
    print("INFO: Adding data matrix...")
    _lx.modes.rna = (
        sparse.csr_matrix(_matrix).transpose(),
        loom_connection.ra["Gene"],
        loom_connection.ca["CellID"],
    )
    ## Add observation (o) attributes
    ### Add annotations
    try:
        print("INFO: Adding annotations...")
        for annotation in metadata.annotations:
            _lx.modes.rna.o.annotations.add(
                key=annotation.name,
                name=annotation.name,
                value=pd.Series(
                    data=loom_connection.ca[annotation.name],
                    index=_lx.modes.rna.X._observation_names,
                    name=annotation.name,
                ),
                force=force_conversion["annotations"]
                if "annotations" in force_conversion
                else False,
            )
    except BadDTypeException:
        raise Exception(
            "You can force the conversion of the annotations to categorical by setting the parameter `force_conversion={'annotations': True}`"
        )
    ### Add metrics
    try:
        print("INFO: Adding metrics...")
        for metric in metadata.metrics:
            _lx.modes.rna.o.metrics.add(
                key=metric.name,
                name=metric.name,
                value=pd.Series(
                    data=loom_connection.ca[metric.name],
                    index=_lx.modes.rna.X._observation_names,
                    name=metric.name,
                ),
                force=force_conversion["metrics"]
                if "metrics" in force_conversion
                else False,
            )
    except BadDTypeException:
        raise Exception(
            "You can force the conversion of the metrics to numerical by setting the parameter `force_conversion={'metrics': True}`"
        )
    ### Add embeddings
    if "Embeddings_X" in loom_connection.ca and "Embeddings_Y" in loom_connection.ca:
        print("INFO: Adding embeddings...")
        for embedding in metadata.embeddings:
            _embedding_df = pd.DataFrame(
                {
                    "_X": loom_connection.ca["Embeddings_X"][str(embedding.id)],
                    "_Y": loom_connection.ca["Embeddings_Y"][str(embedding.id)],
                },
                index=_lx.modes.rna.X._observation_names,
            )
            _lx.modes.rna.o.embeddings.add(
                key=embedding.name,
                value=_embedding_df,
                name=embedding.name,
                metadata=embedding,
            )

    if "Embedding" in loom_connection.ca:
        print("INFO: Adding default embedding...")
        embedding = list(filter(lambda x: int(x.id) == -1, metadata.embeddings))
        if len(embedding) == 0:
            raise Exception(
                "Corrupted LoomX MetaData: missing metadata about default embedding."
            )
        embedding: Embedding
        embedding = embedding[0]
        _lx.modes.rna.o.embeddings.add(
            key=embedding.name,
            value=pd.DataFrame(
                loom_connection.ca["Embedding"],
                index=_lx.modes.rna.X._observation_names,
            ),
            name=embedding.name,
            metadata=embedding,
        )

    ## Add clusterings
    if "Clusterings" in loom_connection.ca:
        print("INFO: Adding clusterings...")
        for clustering in metadata.clusterings:
            _lx.modes.rna.o.clusterings.add(
                key=clustering.name,
                name=clustering.name,
                value=pd.Series(
                    data=loom_connection.ca["Clusterings"][str(clustering.id)],
                    index=_lx.modes.rna.X._observation_names,
                    name=str(clustering.id),
                ),
                metadata=clustering,
            )

    # # Add cluster markers
    _has_markers = False
    clustering_attr: LoomXMetadataV1Clustering
    for _, clustering_attr in _lx.modes.rna.o.clusterings:
        # Check if any markers exist for the current clustering
        cluster_marker_ra_prefix = f"ClusterMarkers_{clustering_attr.id}"
        if cluster_marker_ra_prefix not in loom_connection.ra:
            continue

        print(
            f"INFO: Adding cluster markers for clustering {clustering_attr.name} ({clustering_attr.id})"
        )

        _has_markers = True

        # Make markers table in long format (cluster, <metrics>)
        markers_df = (
            functools.reduce(
                lambda left, right: pd.merge(
                    left, right, on=["index", "variable"], how="left"
                ),
                [
                    pd.DataFrame(
                        loom_connection.ra[cluster_marker_ra_prefix],
                        index=loom_connection.ra.Gene,
                    )
                    .reset_index()
                    .melt("index")
                    .rename(columns={"value": "is_marker"})
                ]
                + [
                    pd.DataFrame(
                        loom_connection.ra[
                            f"{cluster_marker_ra_prefix}_{metric.accessor}"
                        ],
                        index=loom_connection.ra.Gene,
                    )
                    .reset_index()
                    .melt("index")
                    .rename(columns={"value": metric.accessor})
                    for metric in clustering_attr.clusterMarkerMetrics
                ],
            )
            .set_index("index")
            .rename(columns={"variable": "cluster"})
            .query("is_marker == 1")
            .drop(columns=["is_marker"])
        )
        cluster: LoomXMetadataV1Cluster
        for cluster in clustering_attr.clusters:
            cluster.markers = markers_df.query(f"cluster == '{cluster.id}'")

    if not _has_markers:
        print("INFO: No markers detected in the loom.")
    return _lx


def _read_scope_rna_loom(loom_connection: lp.LoomConnection, force_conversion: Dict):
    # Read the Metadata from SCope
    _version, _metadata = _read_scope_metadata(loom_connection=loom_connection)
    if _version == "v1":
        _lx = _read_scope_v1_rna_loom(
            loom_connection=loom_connection,
            metadata=_metadata,
            force_conversion=force_conversion,
        )
    else:
        raise Exception("Unable to detect the LoomX version.")

    print("INFO: Adding global attributes...")
    for global_attr_key in loom_connection.attrs:
        if global_attr_key in GLOBAL_ATTRIBUTE_KEY_VX:
            continue
        _lx.modes.rna.g[global_attr_key] = loom_connection.attrs[global_attr_key]
    print("INFO: Done.")
    return _lx


def _read_scope_loom(
    loom_connection: lp.LoomConnection, mode_type: ModeType, force_conversion: Dict
):
    if mode_type == ModeType.RNA:
        return _read_scope_rna_loom(
            loom_connection=loom_connection, force_conversion=force_conversion
        )
    # Should be uncommented on production
    # Commented here ease the flow with editable mode
    # raise Exception(f"The given mode type '{mode_type}' has not been implemented yet.")


def read_loom(
    file_path: str,
    mode_type: str = "rna",
    force_conversion={"annotations": False, "metrics": False},
) -> LoomX:

    try:
        _mode_type = ModeType(mode_type)
    except:
        mode_types = list(filter(lambda x: x != "_", [w.value for w in (ModeType)]))
        raise Exception(
            f"The given mode type '{mode_type}' does not exist. Choose one of: {', '.join(mode_types)}."
        )

    with lp.connect(filename=file_path, mode="r", validate=False) as loom_connection:
        if any(
            list(map(lambda x: x in loom_connection.attrs, GLOBAL_ATTRIBUTE_KEY_VX))
        ):
            return _read_scope_loom(
                loom_connection=loom_connection,
                mode_type=_mode_type,
                force_conversion=force_conversion,
            )
        raise Exception(f"Unable to read the loom at {file_path}")
